"""
Alerts appear in an activity below the toolbox and above the canvas.

:class:`Alert` and the derived :class:`TimeoutAlert`,
:class:`ConfirmationAlert`, :class:`ErrorAlert`, and
:class:`NotifyAlert`, each have a title, a message and optional
buttons.

:class:`Alert` will emit a `response` signal when a button is
clicked.

The :class:`TimeoutAlert` and :class:`NotifyAlert` display a countdown
and will emit a `response` signal when a timeout occurs.

Example:
    Create a simple alert message.

    .. code-block:: python

        from sugar3.graphics.alert import Alert

        # Create a new simple alert
        alert = Alert()

        # Set the title and text body of the alert
        alert.props.title = _('Title of Alert Goes Here')
        alert.props.msg = _('Text message of alert goes here')

        # Add the alert to the activity
        self.add_alert(alert)
        alert.show()

STABLE.
"""
# Copyright (C) 2007, One Laptop Per Child
# Copyright (C) 2010, Anish Mangal <anishmangal2002@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

import gettext

from gi.repository import Gtk
from gi.repository import GObject
from gi.repository import GLib
import math

from sugar3.graphics import style
from sugar3.graphics.icon import Icon


def _(msg):
    return gettext.dgettext('sugar-toolkit-gtk3', msg)


if not hasattr(GObject.ParamFlags, 'READWRITE'):
    GObject.ParamFlags.READWRITE = GObject.ParamFlags.WRITABLE | \
        GObject.ParamFlags.READABLE


class Alert(Gtk.EventBox):
    """
    Alerts are inside the activity window instead of being a
    separate popup window. They do not hide the canvas.

    Use :func:`~sugar3.graphics.window.Window.add_alert` and
    :func:`~sugar3.graphics.window.Window.remove_alert` to add and
    remove an alert.  These methods are inherited by an
    :class:`~sugar3.activity.activity.Activity` via superclass
    :class:`~sugar3.graphics.window.Window`.

    The alert is placed between the canvas and the toolbox, or above
    the canvas in fullscreen mode.

    Args:
        title (str): the title of the alert
        message (str): the message of the alert
        icon (str): the icon that appears at the far left
    """
    __gtype_name__ = 'SugarAlert'

    __gsignals__ = {
        'response': (GObject.SignalFlags.RUN_FIRST, None, ([object])),
    }

    __gproperties__ = {
        'title': (str, None, None, None, GObject.ParamFlags.READWRITE),
        'msg': (str, None, None, None, GObject.ParamFlags.READWRITE),
        'icon': (object, None, None, GObject.ParamFlags.WRITABLE),
    }

    def __init__(self, **kwargs):
        self._title = None
        self._msg = None
        self._icon = None
        self._buttons = {}

        self._hbox = Gtk.HBox()
        self._hbox.set_border_width(style.DEFAULT_SPACING)
        self._hbox.set_spacing(style.DEFAULT_SPACING)

        self._msg_box = Gtk.VBox()
        self._title_label = Gtk.Label()
        self._title_label.set_alignment(0, 0.5)
        self._title_label.set_ellipsize(style.ELLIPSIZE_MODE_DEFAULT)
        self._msg_box.pack_start(self._title_label, False, False, 0)

        self._msg_label = Gtk.Label()
        self._msg_label.set_alignment(0, 0.5)
        self._msg_label.set_ellipsize(style.ELLIPSIZE_MODE_DEFAULT)
        self._msg_box.pack_start(self._msg_label, False, False, 0)
        self._hbox.pack_start(self._msg_box, False, False, 0)

        self._buttons_box = Gtk.HButtonBox()
        self._buttons_box.set_layout(Gtk.ButtonBoxStyle.END)
        self._buttons_box.set_spacing(style.DEFAULT_SPACING)
        self._hbox.pack_end(self._buttons_box, True, True, 0)

        GObject.GObject.__init__(self, **kwargs)

        self.set_visible_window(True)
        self.add(self._hbox)
        self._title_label.show()
        self._msg_label.show()
        self._buttons_box.show()
        self._msg_box.show()
        self._hbox.show()
        self.show()

    def do_set_property(self, pspec, value):
        """
        Set alert property, GObject internal method.
        Use the `alert.props` object, eg::

            alert.props.title = 'Are you happy?'
        """
        if pspec.name == 'title':
            if self._title != value:
                self._title = value
                self._title_label.set_markup('<b>' + self._title + '</b>')
        elif pspec.name == 'msg':
            if self._msg != value:
                self._msg = value
                self._msg_label.set_markup(self._msg)
                self._msg_label.set_line_wrap(True)
        elif pspec.name == 'icon':
            if self._icon != value:
                self._icon = value
                self._hbox.pack_start(self._icon, False, False, 0)
                self._hbox.reorder_child(self._icon, 0)

    def do_get_property(self, pspec):
        """
        Get alert property, GObject internal method.
        Use the `alert.props` object, eg::

            title = alert.props.title
        """
        if pspec.name == 'title':
            return self._title
        elif pspec.name == 'msg':
            return self._msg

    def add_entry(self):
        """
        Create an entry and add it to the alert.

        The entry is placed after the title and before the buttons.

        Caller is responsible for capturing the entry text in the
        `response` signal handler or a :class:`Gtk.Entry` signal
        handler.

        Returns:
            :class:`Gtk.Entry`: the entry added to the alert
        """
        entry = Gtk.Entry()
        self._hbox.pack_start(entry, True, True, 0)
        entry.show()

        self._hbox.set_child_packing(self._buttons_box, False, False, 0,
                                     Gtk.PackType.END)

        return entry

    def add_button(self, response_id, label, icon=None, position=-1):
        """
        Create a button and add it to the alert.

        The button is added to the end of the alert.

        When the button is clicked, the `response` signal will be
        emitted, along with a response identifier.

        Args:
            response_id (int): the response identifier, a
                :class:`Gtk.ResponseType` constant or any positive
                integer,
            label (str): a label for the button
            icon (:class:`~sugar3.graphics.icon.Icon` or \
                :class:`Gtk.Image`, optional):
                an icon for the button
            position (int, optional): the position of the button in
                the box of buttons,

        Returns:
            :class:`Gtk.Button`: the button added to the alert

        """
        button = Gtk.Button()
        self._buttons[response_id] = button
        if icon is not None:
            button.set_image(icon)
        button.set_label(label)
        self._buttons_box.pack_start(button, True, True, 0)
        button.show()
        button.connect('clicked', self.__button_clicked_cb, response_id)
        if position != -1:
            self._buttons_box.reorder_child(button, position)
        return button

    def remove_button(self, response_id):
        """
        Remove a button from the alert.

        The button is selected for removal using the response
        identifier that was passed to :func:`add_button`.

        Args:
            response_id (int): the response identifier

        Returns:
            None

        """
        self._buttons_box.remove(self._buttons[response_id])

    def _response(self, response_id):
        """
        Emitting response when we have a result

        A result can be that a user has clicked a button or
        a timeout has occurred, the id identifies the button
        that has been clicked and -1 for a timeout
        """
        self.emit('response', response_id)

    def __button_clicked_cb(self, button, response_id):
        self._response(response_id)


if hasattr(Alert, 'set_css_name'):
    Alert.set_css_name('alert')


class ConfirmationAlert(Alert):
    """
    An alert with two buttons; Ok and Cancel.

    When a button is clicked, the :class:`ConfirmationAlert` will emit
    a `response` signal with a response identifier.  For the Ok
    button, the response identifier will be
    :class:`Gtk.ResponseType.OK`.  For the Cancel button,
    :class:`Gtk.ResponseType.CANCEL`.

    Args:
        **kwargs: parameters for :class:`~sugar3.graphics.alert.Alert`

    .. code-block:: python

        from sugar3.graphics.alert import ConfirmationAlert

        # Create a Confirmation alert and add it to the UI.
        def _alert_confirmation(self):
            alert = ConfirmationAlert()
            alert.props.title=_('Title of Alert Goes Here')
            alert.props.msg = _('Text message of alert goes here')
            alert.connect('response', self._alert_response_cb)
            self.add_alert(alert)

        # Called when an alert object sends a response signal.
        def _alert_response_cb(self, alert, response_id):
            # Remove the alert
            self.remove_alert(alert)

            # Check the response identifier.
            if response_id is Gtk.ResponseType.OK:
                print('Ok Button was clicked.')
            elif response_id is Gtk.ResponseType.CANCEL:
                print('Cancel Button was clicked.')
    """

    def __init__(self, **kwargs):
        Alert.__init__(self, **kwargs)

        icon = Icon(icon_name='dialog-cancel')
        self.add_button(Gtk.ResponseType.CANCEL, _('Cancel'), icon)
        icon.show()

        icon = Icon(icon_name='dialog-ok')
        self.add_button(Gtk.ResponseType.OK, _('Ok'), icon)
        icon.show()


class ErrorAlert(Alert):
    """
    An alert with one button; Ok.

    When the button is clicked, the :class:`ErrorAlert` will
    emit a `response` signal with a response identifier
    :class:`Gtk.ResponseType.OK`.

    Args:
        **kwargs: parameters for :class:`~sugar3.graphics.alert.Alert`

    .. code-block:: python

        from sugar3.graphics.alert import ErrorAlert

        # Create a Error alert and add it to the UI.
        def _alert_error(self):
            alert = ErrorAlert()
            alert.props.title=_('Title of Alert Goes Here')
            alert.props.msg = _('Text message of alert goes here')
            alert.connect('response', self._alert_response_cb)
            self.add_alert(alert)

        # called when an alert object throws a response event.
        def _alert_response_cb(self, alert, response_id):
            # Remove the alert
            self.remove_alert(alert)

            # Check the response identifier.
            if response_id is Gtk.ResponseType.OK:
                print('Ok Button was clicked.')
    """

    def __init__(self, **kwargs):
        Alert.__init__(self, **kwargs)

        icon = Icon(icon_name='dialog-ok')
        self.add_button(Gtk.ResponseType.OK, _('Ok'), icon)
        icon.show()


class _TimeoutIcon(Gtk.Alignment):
    __gtype_name__ = 'SugarTimeoutIcon'

    def __init__(self):
        Gtk.Alignment.__init__(self, xalign=0, yalign=0, xscale=1, yscale=1)
        self.set_app_paintable(True)
        self._text = Gtk.Label()
        self._text.set_alignment(0.5, 0.5)
        self.add(self._text)
        self._text.show()
        self.connect('draw', self.__draw_cb)

    def __draw_cb(self, widget, context):
        self._draw(context)
        return False

    def do_get_preferred_width(self):
        width = Gtk.icon_size_lookup(Gtk.IconSize.BUTTON)[1]
        return width, width

    def do_get_preferred_height(self):
        height = Gtk.icon_size_lookup(Gtk.IconSize.BUTTON)[2]
        return height, height

    def _draw(self, context):
        w = self.get_allocated_width()
        h = self.get_allocated_height()
        x = w * 0.5
        y = h * 0.5
        radius = w / 2
        context.arc(x, y, radius, 0, 2 * math.pi)
        widget_style = self.get_style_context()
        color = widget_style.get_background_color(self.get_state_flags())
        context.set_source_rgb(color.red, color.green, color.blue)
        context.fill_preserve()

    def set_text(self, text):
        self._text.set_markup('<b>%s</b>' % GLib.markup_escape_text(str(text)))


if hasattr(_TimeoutIcon, 'set_css_name'):
    _TimeoutIcon.set_css_name('timeouticon')


class _TimeoutAlert(Alert):
    def __init__(self, timeout=5, label=_('Ok'), **kwargs):
        Alert.__init__(self, **kwargs)

        self._timeout = timeout

        self._timeout_text = _TimeoutIcon()
        self._timeout_text.set_text(self._timeout)
        self.add_button(Gtk.ResponseType.OK, label, self._timeout_text)
        self._timeout_text.show()

        self._timeout_sid = GLib.timeout_add(1000, self.__timeout_cb)

    def __timeout_cb(self):
        self._timeout -= 1
        self._timeout_text.set_text(self._timeout)
        if self._timeout == 0:
            Alert._response(self, -1)
            return False
        return True

    def _response(self, *args):
        GLib.source_remove(self._timeout_sid)
        Alert._response(self, *args)


class TimeoutAlert(_TimeoutAlert):
    """
    A timed alert with two buttons; Continue and Cancel.  The Continue
    button contains a countdown of seconds remaining.

    When a button is clicked, the :class:`TimeoutAlert` will emit
    a `response` signal with a response identifier.  For the Continue
    button, the response identifier will be
    :class:`Gtk.ResponseType.OK`.  For the Cancel button,
    :class:`Gtk.ResponseType.CANCEL`.

    If the countdown reaches zero before a button is clicked, the
    :class:`TimeoutAlert` will emit a `response` signal with a
    response identifier of -1.

    Args:
        timeout (int, optional): time in seconds, default 5
        **kwargs: parameters for :class:`~sugar3.graphics.alert.Alert`

    .. code-block:: python

        from sugar3.graphics.alert import TimeoutAlert

        # Create a Timeout alert and add it to the UI
        def _alert_timeout(self):
            alert = TimeoutAlert(timeout=10)
            alert.props.title = _('Title of Alert Goes Here')
            alert.props.msg = _('Text message of alert goes here')
            alert.connect('response', self.__alert_response_cb)
            self.add_alert(alert)

        # Called when an alert object throws a response event.
        def __alert_response_cb(self, alert, response_id):
            # Remove the alert
            self.remove_alert(alert)

            # Check the response identifier.
            if response_id is Gtk.ResponseType.OK:
                print('Continue Button was clicked.')
            elif response_id is Gtk.ResponseType.CANCEL:
                print('Cancel Button was clicked.')
            elif response_id == -1:
                print('Timeout occurred')
    """

    def __init__(self, timeout=5, **kwargs):
        _TimeoutAlert.__init__(self, timeout, _('Continue'), **kwargs)

        icon = Icon(icon_name='dialog-cancel')
        self.add_button(Gtk.ResponseType.CANCEL, _('Cancel'), icon)
        icon.show()


class NotifyAlert(_TimeoutAlert):
    """
    A timed alert with one button; Ok.  The button contains a
    countdown of seconds remaining.

    When the button is clicked, the :class:`NotifyAlert` will
    emit a `response` signal with a response identifier
    :class:`Gtk.ResponseType.OK`.

    If the countdown reaches zero before the button is clicked, the
    :class:`NotifyAlert` will emit a `response` signal with a
    response identifier of -1.

    Args:
        timeout (int, optional): time in seconds, default 5
        **kwargs: parameters for :class:`~sugar3.graphics.alert.Alert`

    .. code-block:: python

        from sugar3.graphics.alert import NotifyAlert

        # create a Notify alert then show it
        def _alert_notify(self):
            alert = NotifyAlert()
            alert.props.title = _('Title of Alert Goes Here')
            alert.props.msg = _('Text message of alert goes here')
            alert.connect('response', self._alert_response_cb)
            self.add_alert(alert)

        def __alert_response_cb(self, alert, response_id):
            # Remove the alert
            self.remove_alert(alert)

            # Check the response identifier.
            if response_id is Gtk.ResponseType.OK:
                print('Ok Button was clicked.')
            elif response_id == -1:
                print('Timeout occurred')
    """

    def __init__(self, timeout=5, **kwargs):
        _TimeoutAlert.__init__(self, timeout, _('Ok'), **kwargs)



# Copyright (C) 2006-2007 Red Hat, Inc.
# Copyright (C) 2010 Collabora Ltd. <http://www.collabora.co.uk/>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""Shell side object which manages request to start activity

UNSTABLE. Activities are currently not allowed to run other activities so at
the moment there is no reason to stabilize this API.
"""

import logging
import dbus
from gi.repository import GObject
from gi.repository import GLib

from sugar3.activity.activityhandle import ActivityHandle
from sugar3 import util
from sugar3 import env
from sugar3.datastore import datastore

from errno import EEXIST, ENOSPC

import os
import subprocess

_SHELL_SERVICE = 'org.laptop.Shell'
_SHELL_PATH = '/org/laptop/Shell'
_SHELL_IFACE = 'org.laptop.Shell'

_ACTIVITY_FACTORY_INTERFACE = 'org.laptop.ActivityFactory'

# helper method to close all filedescriptors
# borrowed from subprocess.py
try:
    MAXFD = os.sysconf('SC_OPEN_MAX')
except ValueError:
    MAXFD = 256


def _close_fds():
    for i in range(3, MAXFD):
        try:
            os.close(i)
        # pylint: disable=W0704
        except Exception:
            pass


def create_activity_id():
    """Generate a new, unique ID for this activity"""
    return util.unique_id()


def _mkdir(path):
    try:
        os.mkdir(path)
    except OSError as e:
        if e.errno != EEXIST:
            raise e


def get_environment(activity):
    environ = os.environ.copy()

    bin_path = os.path.join(activity.get_path(), 'bin')

    activity_root = env.get_profile_path(activity.get_bundle_id())
    _mkdir(activity_root)

    instance_dir = os.path.join(activity_root, 'instance')
    _mkdir(instance_dir)

    data_dir = os.path.join(activity_root, 'data')
    _mkdir(data_dir)

    tmp_dir = os.path.join(activity_root, 'tmp')
    _mkdir(tmp_dir)

    environ['SUGAR_BUNDLE_PATH'] = activity.get_path()
    environ['SUGAR_BUNDLE_ID'] = activity.get_bundle_id()
    environ['SUGAR_ACTIVITY_ROOT'] = activity_root
    environ['PATH'] = bin_path + ':' + environ['PATH']

    if activity.get_path().startswith(env.get_user_activities_path()):
        environ['SUGAR_LOCALEDIR'] = os.path.join(activity.get_path(),
                                                  'locale')

    return environ


def get_command(activity, activity_id=None, object_id=None, uri=None,
                activity_invite=False):
    if not activity_id:
        activity_id = create_activity_id()

    command = activity.get_command().split(' ')
    command.extend(['-b', activity.get_bundle_id()])
    command.extend(['-a', activity_id])

    if object_id is not None:
        command.extend(['-o', object_id])
    if uri is not None:
        command.extend(['-u', uri])
    if activity_invite:
        command.append('-i')

    # if the command is in $BUNDLE_ROOT/bin, execute the absolute path so there
    # is no need to mangle with the shell's PATH
    if '/' not in command[0]:
        bin_path = os.path.join(activity.get_path(), 'bin')
        absolute_path = os.path.join(bin_path, command[0])
        if os.path.exists(absolute_path):
            command[0] = absolute_path

    logging.debug('launching: %r' % command)

    return command


def open_log_file(activity):
    i = 1
    while True:
        path = env.get_logs_path('%s-%s.log' % (activity.get_bundle_id(), i))
        try:
            fd = os.open(path, os.O_EXCL | os.O_CREAT | os.O_WRONLY, 0o644)
            f = os.fdopen(fd, 'w')
            return (path, f)
        except OSError as e:
            if e.errno == EEXIST:
                i += 1
            elif e.errno == ENOSPC:
                # not the end of the world; let's try to keep going.
                return ('/dev/null', open('/dev/null', 'w'))
            else:
                raise e


class ActivityCreationHandler(GObject.GObject):
    """Sugar-side activity creation interface

    This object uses a dbus method on the ActivityFactory
    service to create the new activity.  It generates
    GObject events in response to the success/failure of
    activity startup using callbacks to the service's
    create call.
    """

    def __init__(self, bundle, handle):
        """Initialise the handler

        bundle -- the ActivityBundle to launch
        activity_handle -- stores the values which are to
            be passed to the service to uniquely identify
            the activity to be created and the sharing
            service that may or may not be connected with it

            sugar3.activity.activityhandle.ActivityHandle instance

        calls the "create" method on the service for this
        particular activity type and registers the
        _reply_handler and _error_handler methods on that
        call's results.

        The specific service which creates new instances of this
        particular type of activity is created during the activity
        registration process in shell bundle registry which creates
        service definition files for each registered bundle type.

        If the file '/etc/olpc-security' exists, then activity launching
        will be delegated to the prototype 'Rainbow' security service.
        """
        GObject.GObject.__init__(self)

        self._bundle = bundle
        self._service_name = bundle.get_bundle_id()
        self._handle = handle

        bus = dbus.SessionBus()
        bus_object = bus.get_object(_SHELL_SERVICE, _SHELL_PATH)
        self._shell = dbus.Interface(bus_object, _SHELL_IFACE)

        if handle.activity_id is not None and handle.object_id is None:
            datastore.find({'activity_id': self._handle.activity_id},
                           reply_handler=self._find_object_reply_handler,
                           error_handler=self._find_object_error_handler)
        else:
            self._launch_activity()

    def _launch_activity(self):
        if self._handle.activity_id is not None:
            self._shell.ActivateActivity(
                self._handle.activity_id,
                reply_handler=self._activate_reply_handler,
                error_handler=self._activate_error_handler)
        else:
            self._create_activity()

    def _create_activity(self):
        if self._handle.activity_id is None:
            self._handle.activity_id = create_activity_id()

        self._shell.NotifyLaunch(
            self._service_name, self._handle.activity_id,
            reply_handler=self._no_reply_handler,
            error_handler=self._notify_launch_error_handler)

        environ = get_environment(self._bundle)
        (log_path, log_file) = open_log_file(self._bundle)
        command = get_command(self._bundle, self._handle.activity_id,
                              self._handle.object_id, self._handle.uri,
                              self._handle.invited)

        dev_null = open('/dev/null', 'r')
        child = subprocess.Popen([str(s) for s in command],
                                 env=environ,
                                 cwd=str(self._bundle.get_path()),
                                 close_fds=True,
                                 stdin=dev_null.fileno(),
                                 stdout=log_file.fileno(),
                                 stderr=log_file.fileno())

        GLib.child_watch_add(child.pid,
                             _child_watch_cb,
                             (log_file, self._handle.activity_id))

    def _no_reply_handler(self, *args):
        pass

    def _notify_launch_failure_error_handler(self, err):
        logging.error('Notify launch failure failed %s' % err)

    def _notify_launch_error_handler(self, err):
        logging.debug('Notify launch failed %s' % err)

    def _activate_reply_handler(self, activated):
        if not activated:
            self._create_activity()

    def _activate_error_handler(self, err):
        logging.error('Activity activation request failed %s' % err)

    def _create_reply_handler(self):
        logging.debug('Activity created %s (%s).' %
                      (self._handle.activity_id, self._service_name))

    def _create_error_handler(self, err):
        logging.error("Couldn't create activity %s (%s): %s" %
                      (self._handle.activity_id, self._service_name, err))
        self._shell.NotifyLaunchFailure(
            self._handle.activity_id, reply_handler=self._no_reply_handler,
            error_handler=self._notify_launch_failure_error_handler)

    def _find_object_reply_handler(self, jobjects, count):
        if count > 0:
            if count > 1:
                logging.debug('Multiple objects has the same activity_id.')
            self._handle.object_id = jobjects[0]['uid']
        self._launch_activity()

    def _find_object_error_handler(self, err):
        logging.error('Datastore find failed %s' % err)
        self._launch_activity()


def create(bundle, activity_handle=None):
    """Create a new activity from its name."""
    if not activity_handle:
        activity_handle = ActivityHandle()
    return ActivityCreationHandler(bundle, activity_handle)


def create_with_uri(bundle, uri):
    """Create a new activity and pass the uri as handle."""
    activity_handle = ActivityHandle(uri=uri)
    return ActivityCreationHandler(bundle, activity_handle)


def create_with_object_id(bundle, object_id):
    """Create a new activity and pass the object id as handle."""
    activity_handle = ActivityHandle(object_id=object_id)
    return ActivityCreationHandler(bundle, activity_handle)


def _child_watch_cb(pid, condition, user_data):
    log_file, activity_id = user_data

    if os.WIFEXITED(condition):
        status = os.WEXITSTATUS(condition)
        signum = None
        if status == 0:
            message = 'Normal successful completion'
        else:
            message = 'Exited with status %s' % status
    elif os.WIFSIGNALED(condition):
        status = None
        signum = os.WTERMSIG(condition)
        message = 'Terminated by signal %s' % signum
    else:
        status = None
        signum = os.WTERMSIG(condition)
        message = 'Undefined status with signal %s' % signum

    try:
        log_file.write(
            '%s, pid %s activity_id %s\n' % (message, pid, activity_id))
    finally:
        log_file.close()

    # try to reap zombies in case SIGCHLD has not been set to SIG_IGN
    try:
        os.waitpid(pid, 0)
    except OSError:
        # SIGCHLD = SIG_IGN, no zombies
        pass

    if status or signum:
        # XXX have to recreate dbus object since we can't reuse
        # ActivityCreationHandler's one, see
        # https://bugs.freedesktop.org/show_bug.cgi?id=23507
        bus = dbus.SessionBus()
        bus_object = bus.get_object(_SHELL_SERVICE, _SHELL_PATH)
        shell = dbus.Interface(bus_object, _SHELL_IFACE)

        def reply_handler_cb(*args):
            pass

        def error_handler_cb(error):
            logging.error('Cannot send NotifyLaunchFailure to the shell')

        # TODO send launching failure but activity could already show
        # main window, see http://bugs.sugarlabs.org/ticket/1447#comment:19
        shell.NotifyLaunchFailure(activity_id,
                                  reply_handler=reply_handler_cb,
                                  error_handler=error_handler_cb)



# Copyright (C) 2007 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""Logging service setup.

STABLE.
"""

import six
import array
import collections
import errno
import logging
import sys
import os
import decorator
import time

from six.moves import reprlib as repr_
from sugar3 import env

# Let's keep this module self contained so that it can be easily
# pasted in external sugar service like the datastore.

# traces function calls, use SUGAR_LOGGER_LEVEL=trace to enable
TRACE = 5
_LEVELS = {
    'error': logging.ERROR,
    'warning': logging.WARNING,
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'trace': TRACE,
    'all': 0,
}
logging.addLevelName(TRACE, 'TRACE')


# DEPRECATED
def get_logs_dir():
    return env.get_logs_path()


def set_level(level):
    if level in _LEVELS:
        logging.getLogger('').setLevel(_LEVELS[level])
        return

    try:
        logging.getLogger('').setLevel(int(level))
    except ValueError:
        logging.warning('Invalid log level: %r' % level)


# pylint: disable-msg=E1101,F0401
def _except_hook(exctype, value, traceback):
    # Attempt to provide verbose IPython tracebacks.
    # Importing IPython is slow, so we import it lazily.
    try:
        from IPython.ultraTB import AutoFormattedTB
        sys.excepthook = AutoFormattedTB(mode='Verbose',
                                         color_scheme='NoColor')
    except ImportError:
        sys.excepthook = sys.__excepthook__

    sys.excepthook(exctype, value, traceback)


def cleanup():
    """Clean up the log directory, moving old logs into a numbered backup
    directory.  We only keep `_MAX_BACKUP_DIRS` of these backup directories
    around; the rest are removed."""
    logs_dir = get_logs_dir()

    if not os.path.isdir(logs_dir):
        os.makedirs(logs_dir)

    backup_logs = []
    backup_dirs = []
    for f in os.listdir(logs_dir):
        path = os.path.join(logs_dir, f)
        if os.path.isfile(path):
            backup_logs.append(f)
        elif os.path.isdir(path):
            backup_dirs.append(path)

    if len(backup_dirs) > 3:
        backup_dirs.sort()
        root = backup_dirs[0]

        try:
            for f in os.listdir(root):
                os.remove(os.path.join(root, f))
            os.rmdir(root)
        except OSError as e:
            print("Could not remove old logs files %s" % e)

    if len(backup_logs) > 0:
        name = str(int(time.time()))
        backup_dir = os.path.join(logs_dir, name)
        try:
            os.mkdir(backup_dir)
            for log in backup_logs:
                source_path = os.path.join(logs_dir, log)
                dest_path = os.path.join(backup_dir, log)
                os.rename(source_path, dest_path)
        except OSError as e:
            # gracefully deal w/ disk full
            if e.errno != errno.ENOSPC:
                raise e


def start(log_filename=None):
    logs_path = env.get_logs_path()

    try:
        os.makedirs(logs_path)
    except OSError:
        pass

    # remove existing handlers, or logging.basicConfig() won't have no effect.
    root_logger = logging.getLogger('')
    for handler in root_logger.handlers:
        root_logger.removeHandler(handler)

    class SafeLogWrapper(object):
        """Small file-like wrapper to gracefully handle ENOSPC errors when
        logging."""

        def __init__(self, stream):
            self._stream = stream

        def write(self, s):
            try:
                self._stream.write(s)
            except IOError as e:
                # gracefully deal w/ disk full
                if e.errno != errno.ENOSPC:
                    raise e

        def flush(self):
            try:
                self._stream.flush()
            except IOError as e:
                # gracefully deal w/ disk full
                if e.errno != errno.ENOSPC:
                    raise e

    logging.basicConfig(
        level=logging.WARNING,
        format="%(created)f %(levelname)s %(name)s: %(message)s",
        stream=SafeLogWrapper(sys.stderr))

    if 'SUGAR_LOGGER_LEVEL' in os.environ:
        set_level(os.environ['SUGAR_LOGGER_LEVEL'])

    if log_filename:
        try:
            log_path = os.path.join(logs_path, log_filename + '.log')

            log_fd = os.open(log_path, os.O_WRONLY | os.O_CREAT)
            os.dup2(log_fd, sys.stdout.fileno())
            os.dup2(log_fd, sys.stderr.fileno())
            os.close(log_fd)

            sys.stdout = SafeLogWrapper(sys.stdout)
            sys.stderr = SafeLogWrapper(sys.stderr)
        except OSError as e:
            # if we're out of space, just continue
            if e.errno != errno.ENOSPC:
                raise e

    sys.excepthook = _except_hook


class TraceRepr(repr_.Repr):

    # better handling of subclasses of basic types, e.g. for DBus
    _TYPES = [int, bool, tuple, list, array.array, set, frozenset,
              collections.deque, dict, str]
    if six.PY2:
        _TYPES.append(long)

    def repr1(self, x, level):
        for t in self._TYPES:
            if isinstance(x, t):
                return getattr(self, 'repr_' + t.__name__)(x, level)

        return repr_.Repr.repr1(self, x, level)

    def repr_int(self, x, level):
        return repr(x)

    def repr_bool(self, x, level):
        return repr(x)


def trace(logger=None, logger_name=None, skip_args=None, skip_kwargs=None,
          maxsize_list=30, maxsize_dict=30, maxsize_string=300):

    if skip_args is None:
        skip_args = []

    if skip_kwargs is None:
        skip_kwargs = []

    # size-limit repr()
    trace_repr = TraceRepr()
    trace_repr.maxlist = maxsize_list
    trace_repr.maxdict = maxsize_dict
    trace_repr.maxstring = maxsize_string
    trace_repr.maxother = maxsize_string
    trace_logger = logger or logging.getLogger(logger_name)

    def _trace(f, *args, **kwargs):
        # don't do expensive formatting if loglevel TRACE is not enabled
        enabled = trace_logger.isEnabledFor(TRACE)
        if not enabled:
            return f(*args, **kwargs)

        params_formatted = ", ".join(
            [trace_repr.repr(a)
                for (idx, a) in enumerate(args) if idx not in skip_args] +
            ['%s=%s' % (k, trace_repr.repr(v))
                for (k, v) in list(kwargs.items()) if k not in skip_kwargs])

        trace_logger.log(TRACE, "%s(%s) invoked", f.__name__,
                         params_formatted)

        try:
            res = f(*args, **kwargs)
        except BaseException:
            trace_logger.exception("Exception occurred in %s" % f.__name__)
            raise

        trace_logger.log(TRACE, "%s(%s) returned %s", f.__name__,
                         params_formatted, trace_repr.repr(res))

        return res

    return decorator.decorator(_trace)



# Copyright (C) 2012, Daniel Narvaez
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
UNSTABLE.
"""

import logging
import os
import unittest
import subprocess
from contextlib import contextmanager

from sugar3.test import uitree

import dbus
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)
bus = dbus.SessionBus()


class UITestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(UITestCase, self).__init__(*args, **kwargs)
        self.bundle_id = None

    def setUp(self):
        logger = logging.getLogger()
        self._orig_level = logger.getEffectiveLevel()
        logger.setLevel(logging.DEBUG)

    def tearDown(self):
        logger = logging.getLogger()
        logger.setLevel(self._orig_level)

    @contextmanager
    def run_view(self, name):
        view_path = os.path.join("views", "%s.py" % name)
        process = subprocess.Popen(["python3", view_path])

        try:
            yield
        except BaseException:
            logging.debug(uitree.get_root().dump())
            raise
        finally:
            process.terminate()

    @contextmanager
    def run_activity(self, options=None):
        self._run_activity(options)

    @contextmanager
    def run_activity_with_uri(self, uri):
        self._run_activity(['--uri', uri])

    def _run_activity(self, options=None):
        if self.bundle_id is not None:
            cmd = ["sugar-launch", self.bundle_id]
            if options is not None:
                cmd += options
            process = subprocess.Popen(cmd)
        else:
            print("No bundle_id specified.")
            return

        try:
            yield
        except BaseException:
            logging.debug(uitree.get_root().dump())
            raise
        finally:
            process.terminate()



# Copyright (C) 2013, One Laptop Per Child
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
'''
A progress icon is a progress indicator in the form of an icon.
'''

from gi.repository import Gtk
from sugar3.graphics.icon import get_surface
from sugar3.graphics import style


class ProgressIcon(Gtk.DrawingArea):
    '''
    Display the progress filling the icon.
    This class is compatible with the sugar3.graphics.icon.Icon class.
    Call update(progress) with the new progress to update the icon.
    The direction defaults to 'vertical', in which case the icon is
    filled from bottom to top.  If direction is set to 'horizontal',
    it will be filled from right to left or from left to right,
    depending on the system's language RTL setting.
    Args:
      pixel_size (integer): sets the icon size
         [e.g. pixel_size=style.LARGE_ICON_SIZE]
      icon_name (string): Name of icon
         [e.g. icon_name='test_icon']
      stroke_color (string): Stroke color means border color.
      fill_color (string): The main (inside) color of progressicon
         [e.g. fill_color=style.COLOR_BLUE.get_svg()
    '''

    def __init__(self, icon_name, pixel_size, stroke_color, fill_color,
                 direction='vertical'):
        Gtk.DrawingArea.__init__(self)

        self._icon_name = icon_name
        self._direction = direction
        self._progress = 0

        self._stroke = get_surface(
            icon_name=icon_name, width=pixel_size, height=pixel_size,
            stroke_color=stroke_color,
            fill_color=style.COLOR_TRANSPARENT.get_svg())

        self._fill = get_surface(
            icon_name=icon_name, width=pixel_size, height=pixel_size,
            stroke_color=style.COLOR_TRANSPARENT.get_svg(),
            fill_color=fill_color)

        self.connect("draw", self.__draw_cb)

    def __draw_cb(self, widget, cr):
        allocation = widget.get_allocation()

        # Center the graphic in the allocated space.
        margin_x = (allocation.width - self._stroke.get_width()) / 2
        margin_y = (allocation.height - self._stroke.get_height()) / 2
        cr.translate(margin_x, margin_y)

        # Paint the fill, clipping it by the progress.
        x_, y_ = 0, 0
        width, height = self._stroke.get_width(), self._stroke.get_height()
        if self._direction == 'vertical':  # vertical direction, bottom to top
            y_ = self._stroke.get_height()
            height *= self._progress * -1
        else:
            rtl_direction = \
                Gtk.Widget.get_default_direction() == Gtk.TextDirection.RTL
            if rtl_direction:  # horizontal direction, right to left
                x_ = self._stroke.get_width()
                width *= self._progress * -1
            else:  # horizontal direction, left to right
                width *= self._progress

        cr.rectangle(x_, y_, width, height)
        cr.clip()
        cr.set_source_surface(self._fill, 0, 0)
        cr.paint()

        # Paint the stroke over the fill.
        cr.reset_clip()
        cr.set_source_surface(self._stroke, 0, 0)
        cr.paint()

    def do_get_preferred_width(self):
        '''
        Calculate the minimum and natural width of the progressicon.
        '''
        width = self._stroke.get_width()
        return (width, width)

    def do_get_preferred_height(self):
        '''
        Calculate the minimum and natural height of the progressicon.
        '''
        height = self._stroke.get_height()
        return (height, height)

    def update(self, progress):
        '''
        Updates progressicon with progress's value.
        Example: update(0.9)
        '''
        self._progress = progress
        self.queue_draw()



# Copyright (C) 2011 One Laptop Per Child
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import logging
from gettext import gettext as _

import gi
gi.require_version('Gtk', '3.0')

from gi.repository import Gio
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject

_HAS_GST = True
try:
    gi.require_version('Gst', '1.0')
    from gi.repository import Gst
    Gst.init(None)
    Gst.parse_launch('espeak')
except BaseException:
    logging.error('Gst or the espeak plugin is not installed in the system.')
    _HAS_GST = False

from sugar3 import power

DEFAULT_PITCH = 0


DEFAULT_RATE = 0


_SAVE_TIMEOUT = 500


SPEECH_SCHEMA = 'org.sugarlabs.speech'


# This voice names are use dto allow the translation of the voice names.
# If espeak add new voices, we need update this list.

translated_espeak_voices = {
    # Translators:  http://en.wikipedia.org/wiki/Afrikaans_language
    'af': _('Afrikaans'),
    # Translators:  http://en.wikipedia.org/wiki/Aragonese_language
    'an': _('Aragonese'),
    # Translators:  http://en.wikipedia.org/wiki/Bulgarian_language
    'bg': _('Bulgarian'),
    # Translators:  http://en.wikipedia.org/wiki/Bosnian_language
    'bs': _('Bosnian'),
    # Translators:  http://en.wikipedia.org/wiki/Catalan_language
    'ca': _('Catalan'),
    # Translators:  http://en.wikipedia.org/wiki/Czech_language
    'cs': _('Czech'),
    # Translators:  http://en.wikipedia.org/wiki/Welsh_language
    'cy': _('Welsh'),
    # Translators:  http://en.wikipedia.org/wiki/Danish_language
    'da': _('Danish'),
    # Translators:  http://en.wikipedia.org/wiki/German_language
    'de': _('German'),
    # Translators:  http://en.wikipedia.org/wiki/Modern_Greek
    'el': _('Greek'),
    'en': _('English'),
    # Translators:  http://en.wikipedia.org/wiki/British_English
    'en_gb': _('English Britain'),
    # Translators:  http://en.wikipedia.org/wiki/Scottish_English
    'en_sc': _('English scottish'),
    'en_uk-north': _('English-north'),
    # Translators:  http://en.wikipedia.org/wiki/Received_Pronunciation
    'en_uk-rp': _('English_rp'),
    # Translators:  http://en.wikipedia.org/wiki/West_Midlands_English
    'en_uk-wmids': _('English_wmids'),
    # Translators:  http://en.wikipedia.org/wiki/American_English
    'en_us': _('English USA'),
    # Translators:  http://en.wikipedia.org/wiki/Caribbean_English
    'en_wi': _('English West Indies'),
    # Translators:  http://en.wikipedia.org/wiki/Esperanto
    'eo': _('Esperanto'),
    # Translators:  http://en.wikipedia.org/wiki/Spanish_language
    'es': _('Spanish'),
    'es_la': _('Spanish latin american'),
    # Translators:  http://en.wikipedia.org/wiki/Estonian_language
    'et': _('Estonian'),
    # Translators:  http://en.wikipedia.org/wiki/Persian_language
    'fa': _('Farsi'),
    # Translators:  http://en.wikipedia.org/wiki/Fingilish
    'fa_pin': _('Farsi-pinglish'),
    # Translators:  http://en.wikipedia.org/wiki/Finnish_language
    'fi': _('Finnish'),
    # Translators:  http://en.wikipedia.org/wiki/Belgian_French
    'fr_be': _('French belgium'),
    # Translators:  http://en.wikipedia.org/wiki/French_language
    'fr_fr': _('French'),
    # Translators:  http://en.wikipedia.org/wiki/Irish_language
    'ga': _('Irish-gaeilge'),
    # Translators:  http://en.wikipedia.org/wiki/Ancient_Greek
    'grc': _('Greek-ancient'),
    # Translators:  http://en.wikipedia.org/wiki/Hindi
    'hi': _('Hindi'),
    # Translators:  http://en.wikipedia.org/wiki/Croatian_language
    'hr': _('Croatian'),
    # Translators:  http://en.wikipedia.org/wiki/Hungarian_language
    'hu': _('Hungarian'),
    # Translators:  http://en.wikipedia.org/wiki/Armenian_language
    'hy': _('Armenian'),
    # Translators:  http://en.wikipedia.org/wiki/Western_Armenian
    'hy_west': _('Armenian (west)'),
    # Translators:  http://en.wikipedia.org/wiki/Indonesian_language
    'id': _('Indonesian'),
    # Translators:  http://en.wikipedia.org/wiki/Icelandic_language
    'is': _('Icelandic'),
    # Translators:  http://en.wikipedia.org/wiki/Italian_language
    'it': _('Italian'),
    # Translators:  http://en.wikipedia.org/wiki/Lojban
    'jbo': _('Lojban'),
    # Translators:  http://en.wikipedia.org/wiki/Georgian_language
    'ka': _('Georgian'),
    # Translators:  http://en.wikipedia.org/wiki/Kannada_language
    'kn': _('Kannada'),
    # Translators:  http://en.wikipedia.org/wiki/Kurdish_language
    'ku': _('Kurdish'),
    # Translators:  http://en.wikipedia.org/wiki/Latin
    'la': _('Latin'),
    # Translators:  http://en.wikipedia.org/wiki/Lithuanian_language
    'lt': _('Lithuanian'),
    # Translators:  http://en.wikipedia.org/wiki/Latvian_language
    'lv': _('Latvian'),
    # Translators:  http://en.wikipedia.org/wiki/Macedonian_language
    'mk': _('Macedonian'),
    # Translators:  http://en.wikipedia.org/wiki/Malayalam
    'ml': _('Malayalam'),
    # Translators:  http://en.wikipedia.org/wiki/Malay_language
    'ms': _('Malay'),
    # Translators:  http://en.wikipedia.org/wiki/Nepali_language
    'ne': _('Nepali'),
    # Translators:  http://en.wikipedia.org/wiki/Dutch_language
    'nl': _('Dutch'),
    # Translators:  http://en.wikipedia.org/wiki/Norwegian_language
    'no': _('Norwegian'),
    # Translators:  http://en.wikipedia.org/wiki/Punjabi_language
    'pa': _('Punjabi'),
    # Translators:  http://en.wikipedia.org/wiki/Polish_language
    'pl': _('Polish'),
    # Translators:  http://en.wikipedia.org/wiki/Brazilian_Portuguese
    'pt_br': _('Portuguese (Brazil)'),
    # Translators:  http://en.wikipedia.org/wiki/Portuguese_language
    'pt_pt': _('Portuguese (Portugal)'),
    # Translators:  http://en.wikipedia.org/wiki/Romanian_language
    'ro': _('Romanian'),
    # Translators:  http://en.wikipedia.org/wiki/Russian_language
    'ru': _('Russian'),
    # Translators:  http://en.wikipedia.org/wiki/Slovak_language
    'sk': _('Slovak'),
    # Translators:  http://en.wikipedia.org/wiki/Albanian_language
    'sq': _('Albanian'),
    # Translators:  http://en.wikipedia.org/wiki/Serbian_language
    'sr': _('Serbian'),
    # Translators:  http://en.wikipedia.org/wiki/Swedish_language
    'sv': _('Swedish'),
    # Translators:  http://en.wikipedia.org/wiki/Swahili_language
    'sw': _('Swahili'),
    # Translators:  http://en.wikipedia.org/wiki/Tamil_language
    'ta': _('Tamil'),
    # Translators:  http://en.wikipedia.org/wiki/Turkish_language
    'tr': _('Turkish'),
    # Translators:  http://en.wikipedia.org/wiki/Vietnamese_language
    'vi': _('Vietnam'),
    'vi_hue': _('Vietnam_hue'),
    'vi_sgn': _('Vietnam_sgn'),
    # Translators:  http://en.wikipedia.org/wiki/Mandarin_Chinese
    'zh': _('Mandarin'),
    # Translators:  http://en.wikipedia.org/wiki/Cantonese
    'zh_yue': _('Cantonese')
}


class SpeechManager(GObject.GObject):

    __gtype_name__ = 'SpeechManager'

    __gsignals__ = {
        'play': (GObject.SignalFlags.RUN_FIRST, None, []),
        'pause': (GObject.SignalFlags.RUN_FIRST, None, []),
        'stop': (GObject.SignalFlags.RUN_FIRST, None, []),
        'mark': (GObject.SignalFlags.RUN_FIRST, None, [str])
    }

    MIN_PITCH = -100
    MAX_PITCH = 100

    MIN_RATE = -100
    MAX_RATE = 100

    def __init__(self, **kwargs):
        GObject.GObject.__init__(self, **kwargs)
        self.player = None
        if not self.enabled():
            return

        self.player = GstSpeechPlayer()
        self.player.connect('play', self._update_state, 'play')
        self.player.connect('stop', self._update_state, 'stop')
        self.player.connect('pause', self._update_state, 'pause')
        self.player.connect('mark', self._mark_cb)
        self._default_voice_name = self.player.get_default_voice()
        self._pitch = DEFAULT_PITCH
        self._rate = DEFAULT_RATE
        self._is_playing = False
        self._is_paused = False
        self._save_timeout_id = -1
        self.restore()

    def enabled(self):
        return _HAS_GST

    def _update_state(self, player, signal):
        self._is_playing = (signal == 'play')
        self._is_paused = (signal == 'pause')
        self.emit(signal)

    def _mark_cb(self, player, value):
        self.emit('mark', value)

    def get_is_playing(self):
        return self._is_playing

    is_playing = GObject.Property(type=bool, getter=get_is_playing,
                                  setter=None, default=False)

    def get_is_paused(self):
        return self._is_paused

    is_paused = GObject.Property(type=bool, getter=get_is_paused,
                                 setter=None, default=False)

    def get_pitch(self):
        return self._pitch

    def get_rate(self):
        return self._rate

    def set_pitch(self, pitch):
        self._pitch = pitch
        if self._save_timeout_id != -1:
            GLib.source_remove(self._save_timeout_id)
        self._save_timeout_id = GLib.timeout_add(_SAVE_TIMEOUT, self.save)

    def set_rate(self, rate):
        self._rate = rate
        if self._save_timeout_id != -1:
            GLib.source_remove(self._save_timeout_id)
        self._save_timeout_id = GLib.timeout_add(_SAVE_TIMEOUT, self.save)

    def say_text(self, text, pitch=None, rate=None, lang_code=None):
        if pitch is None:
            pitch = self._pitch
        if rate is None:
            rate = self._rate
        if lang_code is None:
            voice_name = self._default_voice_name
        else:
            voice_name = self.player.get_all_voices()[lang_code]
        if text:
            logging.debug(
                'PLAYING %r lang %r pitch %r rate %r',
                text,
                voice_name,
                pitch,
                rate)
            self.player.speak(pitch, rate, voice_name, text)

    def say_selected_text(self):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
        clipboard.request_text(self.__primary_selection_cb, None)

    def pause(self):
        self.player.pause_sound_device()

    def restart(self):
        self.player.restart_sound_device()

    def stop(self):
        self.player.stop_sound_device()

    def __primary_selection_cb(self, clipboard, text, user_data):
        self.say_text(text)

    def save(self):
        self._save_timeout_id = -1

        schema_source = Gio.SettingsSchemaSource.get_default()
        if schema_source.lookup(SPEECH_SCHEMA, True) is None:
            return False

        settings = Gio.Settings(SPEECH_SCHEMA)
        settings.set_int('pitch', self._pitch)
        settings.set_int('rate', self._rate)
        logging.debug('saving speech configuration pitch %s rate %s',
                      self._pitch, self._rate)
        return False

    def restore(self):
        schema_source = Gio.SettingsSchemaSource.get_default()
        if schema_source.lookup(SPEECH_SCHEMA, True) is None:
            return

        settings = Gio.Settings(SPEECH_SCHEMA)
        self._pitch = settings.get_int('pitch')
        self._rate = settings.get_int('rate')
        logging.debug('loading speech configuration pitch %s rate %s',
                      self._pitch, self._rate)

    def get_all_voices(self):
        if self.player:
            return self.player.get_all_voices()
        return None

    def get_all_traslated_voices(self):
        """ deprecated after 0.112, due to method name spelling error """
        if self.player:
            return self.player.get_all_translated_voices()
        return None

    def get_all_translated_voices(self):
        if self.player:
            return self.player.get_all_translated_voices()
        return None


class GstSpeechPlayer(GObject.GObject):

    __gsignals__ = {
        'play': (GObject.SignalFlags.RUN_FIRST, None, []),
        'pause': (GObject.SignalFlags.RUN_FIRST, None, []),
        'stop': (GObject.SignalFlags.RUN_FIRST, None, []),
        'mark': (GObject.SignalFlags.RUN_FIRST, None, [str])
    }

    def __init__(self):
        GObject.GObject.__init__(self)
        self.pipeline = None
        self._all_voices = None
        self._all_translated_voices = None

    def restart_sound_device(self):
        if self.pipeline is None:
            logging.debug('Trying to restart not initialized sound device')
            return

        power.get_power_manager().inhibit_suspend()
        self.pipeline.set_state(Gst.State.PLAYING)
        self.emit('play')

    def pause_sound_device(self):
        if self.pipeline is None:
            return

        self.pipeline.set_state(Gst.State.PAUSED)
        power.get_power_manager().restore_suspend()
        self.emit('pause')

    def stop_sound_device(self):
        if self.pipeline is None:
            return

        self.pipeline.set_state(Gst.State.NULL)
        power.get_power_manager().restore_suspend()
        self.emit('stop')

    def make_pipeline(self, command):
        if self.pipeline is not None:
            self.stop_sound_device()
            del self.pipeline

        self.pipeline = Gst.parse_launch(command)

        bus = self.pipeline.get_bus()
        bus.add_signal_watch()
        bus.connect('message', self.__pipe_message_cb)

    def __pipe_message_cb(self, bus, message):
        if message.type in (Gst.MessageType.EOS, Gst.MessageType.ERROR):
            self.pipeline.set_state(Gst.State.NULL)
            self.pipeline = None
            power.get_power_manager().restore_suspend()
            self.emit('stop')
        elif message.type is Gst.MessageType.ELEMENT and \
                message.get_structure().get_name() == 'espeak-mark':
            mark_value = message.get_structure().get_value('mark')
            self.emit('mark', mark_value)

    def speak(self, pitch, rate, voice_name, text):
        # TODO workaround for http://bugs.sugarlabs.org/ticket/1801
        if not [i for i in text if i.isalnum()]:
            return

        self.make_pipeline('espeak name=espeak ! autoaudiosink')
        src = self.pipeline.get_by_name('espeak')

        src.props.text = text
        src.props.pitch = pitch
        src.props.rate = rate
        src.props.voice = voice_name
        src.props.track = 2  # track for marks

        self.restart_sound_device()

    def get_all_voices(self):
        if self._all_voices is not None:
            return self._all_voices
        self._init_voices()
        return self._all_voices

    def get_all_translated_voices(self):
        if self._all_translated_voices is not None:
            return self._all_translated_voices
        self._init_voices()
        return self._all_translated_voices

    def _init_voices(self):
        self._all_voices = {}
        self._all_translated_voices = {}

        for voice in Gst.ElementFactory.make('espeak', None).props.voices:
            name, language, dialect = voice
            if dialect != 'none':
                lang_code = language + '_' + dialect
            else:
                lang_code = language

            self._all_voices[lang_code] = name
            if lang_code in translated_espeak_voices:
                self._all_translated_voices[lang_code] = \
                    translated_espeak_voices[lang_code]
            else:
                self._all_translated_voices[lang_code] = name

    def get_default_voice(self):
        """Try to figure out the default voice, from the current locale ($LANG)
           Fall back to espeak's voice called Default."""
        voices = self.get_all_voices()

        locale = os.environ.get('LANG', '')
        language_location = locale.split('.', 1)[0].lower()
        language = language_location.split('_')[0]
        # if the language is es but not es_es default to es_la (latin voice)
        if language == 'es' and language_location != 'es_es':
            language_location = 'es_la'

        best = voices.get(language_location) or voices.get(language) \
            or 'english'
        return best



# This should eventually land in telepathy-python, so has the same license:

# Copyright (C) 2007 Collabora Ltd. <http://www.collabora.co.uk/>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
STABLE.
"""

__all__ = ('TubeConnection', )
__docformat__ = 'reStructuredText'


import six
import logging

from dbus.connection import Connection


logger = logging.getLogger('telepathy.tubeconn')


class TubeConnection(Connection):

    def __new__(cls, conn, tubes_iface, tube_id, address=None,
                group_iface=None, mainloop=None):
        # pylint: disable=W0212
        # Confused by __new__
        if address is None:
            address = tubes_iface.GetDBusTubeAddress(tube_id)
        self = super(TubeConnection, cls).__new__(cls, address,
                                                  mainloop=mainloop)

        self._tubes_iface = tubes_iface
        self.tube_id = tube_id
        self.participants = {}
        self.bus_name_to_handle = {}
        self._mapping_watches = []

        if group_iface is None:
            method = conn.GetSelfHandle
        else:
            method = group_iface.GetSelfHandle
        method(reply_handler=self._on_get_self_handle_reply,
               error_handler=self._on_get_self_handle_error)

        return self

    def _on_get_self_handle_reply(self, handle):
        # pylint: disable=W0201
        # Confused by __new__
        self.self_handle = handle
        match = self._tubes_iface.connect_to_signal(
            'DBusNamesChanged',
            self._on_dbus_names_changed)
        self._tubes_iface.GetDBusNames(
            self.tube_id,
            reply_handler=self._on_get_dbus_names_reply,
            error_handler=self._on_get_dbus_names_error)
        self._dbus_names_changed_match = match

    def _on_get_self_handle_error(self, e):
        logging.basicConfig()
        logger.error('GetSelfHandle failed: %s', e)

    def close(self):
        self._dbus_names_changed_match.remove()
        self._on_dbus_names_changed(
            self.tube_id, (), list(
                self.participants.keys()))
        super(TubeConnection, self).close()

    def _on_get_dbus_names_reply(self, names):
        self._on_dbus_names_changed(self.tube_id, names, ())

    def _on_get_dbus_names_error(self, e):
        logging.basicConfig()
        logger.error('GetDBusNames failed: %s', e)

    def _on_dbus_names_changed(self, tube_id, added, removed):
        if tube_id == self.tube_id:
            for handle, bus_name in added:
                if handle == self.self_handle:
                    # I've just joined - set my unique name
                    self.set_unique_name(bus_name)
                self.participants[handle] = bus_name
                self.bus_name_to_handle[bus_name] = handle

            # call the callback while the removed people are still in
            # participants, so their bus names are available
            for callback in self._mapping_watches:
                callback(added, removed)

            for handle in removed:
                bus_name = self.participants.pop(handle, None)
                self.bus_name_to_handle.pop(bus_name, None)

    def watch_participants(self, callback):
        self._mapping_watches.append(callback)
        if self.participants:
            # GetDBusNames already returned: fake a participant add event
            # immediately
            added = []
            for k, v in six.iteritems(self.participants):
                added.append((k, v))
            callback(added, [])



# Copyright (C) 2007, Red Hat, Inc.
# Copyright (C) 2010 Collabora Ltd. <http://www.collabora.co.uk/>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""UI interface to an activity in the presence service

STABLE.
"""

import six
import logging
from functools import partial

import gi
gi.require_version('TelepathyGLib', '0.12')
import dbus
from dbus import PROPERTIES_IFACE
from gi.repository import GObject
from gi.repository import TelepathyGLib

from sugar3.presence.buddy import Buddy

CHANNEL = TelepathyGLib.IFACE_CHANNEL
CHANNEL_INTERFACE_GROUP = TelepathyGLib.IFACE_CHANNEL_INTERFACE_GROUP
CONN_INTERFACE_ROOM_CONFIG = \
    TelepathyGLib.IFACE_CHANNEL_INTERFACE_ROOM_CONFIG
CHANNEL_TYPE_TUBES = TelepathyGLib.IFACE_CHANNEL_TYPE_TUBES
CHANNEL_TYPE_TEXT = TelepathyGLib.IFACE_CHANNEL_TYPE_TEXT
CONNECTION = TelepathyGLib.IFACE_CONNECTION
PROPERTIES_INTERFACE = TelepathyGLib.IFACE_PROPERTIES_INTERFACE

CHANNEL_GROUP_FLAG_CHANNEL_SPECIFIC_HANDLES = \
    TelepathyGLib.ChannelGroupFlags.CHANNEL_SPECIFIC_HANDLES
HANDLE_TYPE_CONTACT = TelepathyGLib.HandleType.CONTACT
HANDLE_TYPE_ROOM = TelepathyGLib.HandleType.ROOM
PROPERTY_FLAG_WRITE = TelepathyGLib.PropertyFlags.WRITE

CONN_INTERFACE_ACTIVITY_PROPERTIES = 'org.laptop.Telepathy.ActivityProperties'
CONN_INTERFACE_BUDDY_INFO = 'org.laptop.Telepathy.BuddyInfo'

_logger = logging.getLogger('sugar3.presence.activity')


if not hasattr(GObject.ParamFlags, 'READWRITE'):
    GObject.ParamFlags.READWRITE = GObject.ParamFlags.WRITABLE | \
        GObject.ParamFlags.READABLE


class Activity(GObject.GObject):
    """UI interface for an Activity in the presence service

    Activities in the presence service represent your and other user's
    shared activities.

    Properties:
        id
        color
        name
        type
        joined
    """
    __gsignals__ = {
        'buddy-joined': (GObject.SignalFlags.RUN_FIRST, None,
                         ([GObject.TYPE_PYOBJECT])),
        'buddy-left': (GObject.SignalFlags.RUN_FIRST, None,
                       ([GObject.TYPE_PYOBJECT])),
        'new-channel': (GObject.SignalFlags.RUN_FIRST, None,
                        ([GObject.TYPE_PYOBJECT])),
        'joined': (GObject.SignalFlags.RUN_FIRST, None,
                   ([GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT])),
    }

    __gproperties__ = {
        'id': (str, None, None, None, GObject.ParamFlags.READABLE),
        'name': (str, None, None, None, GObject.ParamFlags.READWRITE),
        'tags': (str, None, None, None, GObject.ParamFlags.READWRITE),
        'color': (str, None, None, None, GObject.ParamFlags.READWRITE),
        'type': (str, None, None, None, GObject.ParamFlags.READABLE),
        'private': (bool, None, None, True, GObject.ParamFlags.READWRITE),
        'joined': (bool, None, None, False, GObject.ParamFlags.READABLE),
    }

    def __init__(self, account_path, connection, room_handle=None,
                 properties=None):
        if room_handle is None and properties is None:
            raise ValueError('Need to pass one of room_handle or properties')

        if properties is None:
            properties = {}

        GObject.GObject.__init__(self)

        self._account_path = account_path
        self.telepathy_conn = connection
        self.telepathy_text_chan = None
        self.telepathy_tubes_chan = None

        self.room_handle = room_handle
        self._join_command = None
        self._share_command = None
        self._id = properties.get('id', None)
        self._color = properties.get('color', None)
        self._name = properties.get('name', None)
        self._type = properties.get('type', None)
        self._tags = properties.get('tags', None)
        self._private = properties.get('private', True)
        self._joined = properties.get('joined', False)
        self._channel_self_handle = None
        self._text_channel_group_flags = 0
        self._buddies = {}
        self._joined_buddies = {}

        self._get_properties_call = None
        if self.room_handle is not None:
            self._start_tracking_properties()

    def _start_tracking_properties(self):
        bus = dbus.SessionBus()
        if six.PY2:
            self._get_properties_call = bus.call_async(
                self.telepathy_conn.requested_bus_name,
                self.telepathy_conn.object_path,
                CONN_INTERFACE_ACTIVITY_PROPERTIES,
                'GetProperties',
                'u',
                (self.room_handle,),
                reply_handler=self.__got_properties_cb,
                error_handler=self.__error_handler_cb,
                utf8_strings=True)
        else:
            self._get_properties_call = bus.call_async(
                self.telepathy_conn.requested_bus_name,
                self.telepathy_conn.object_path,
                CONN_INTERFACE_ACTIVITY_PROPERTIES,
                'GetProperties',
                'u',
                (self.room_handle,),
                reply_handler=self.__got_properties_cb,
                error_handler=self.__error_handler_cb)

        # As only one Activity instance is needed per activity process,
        # we can afford listening to ActivityPropertiesChanged like this.
        self.telepathy_conn.connect_to_signal(
            'ActivityPropertiesChanged',
            self.__activity_properties_changed_cb,
            dbus_interface=CONN_INTERFACE_ACTIVITY_PROPERTIES)

    def __activity_properties_changed_cb(self, room_handle, properties):
        _logger.debug('%r: Activity properties changed to %r' % (self,
                                                                 properties))
        self._update_properties(properties)

    def __got_properties_cb(self, properties):
        _logger.debug('__got_properties_cb %r' % properties)
        self._get_properties_call = None
        self._update_properties(properties)

    def __error_handler_cb(self, error):
        _logger.debug('__error_handler_cb %r' % error)

    def _update_properties(self, new_props):
        val = new_props.get('name', self._name)
        if isinstance(val, str) and val != self._name:
            self._name = val
            self.notify('name')
        val = new_props.get('tags', self._tags)
        if isinstance(val, str) and val != self._tags:
            self._tags = val
            self.notify('tags')
        val = new_props.get('color', self._color)
        if isinstance(val, str) and val != self._color:
            self._color = val
            self.notify('color')
        val = bool(new_props.get('private', self._private))
        if val != self._private:
            self._private = val
            self.notify('private')
        val = new_props.get('id', self._id)
        if isinstance(val, str) and self._id is None:
            self._id = val
            self.notify('id')
        val = new_props.get('type', self._type)
        if isinstance(val, str) and self._type is None:
            self._type = val
            self.notify('type')

    def object_path(self):
        """Get our dbus object path"""
        return self._object_path

    def do_get_property(self, pspec):
        """Retrieve a particular property from our property dictionary"""

        if pspec.name == 'joined':
            return self._joined

        if self._get_properties_call is not None:
            _logger.debug('%r: Blocking on GetProperties() because someone '
                          'wants property %s' % (self, pspec.name))
            self._get_properties_call.block()

        if pspec.name == 'id':
            return self._id
        elif pspec.name == 'name':
            return self._name
        elif pspec.name == 'color':
            return self._color
        elif pspec.name == 'type':
            return self._type
        elif pspec.name == 'tags':
            return self._tags
        elif pspec.name == 'private':
            return self._private

    def do_set_property(self, pspec, val):
        """Set a particular property in our property dictionary"""
        # FIXME: need an asynchronous API to set these properties,
        # particularly 'private'

        if pspec.name == 'name':
            self._name = val
        elif pspec.name == 'color':
            self._color = val
        elif pspec.name == 'tags':
            self._tags = val
        elif pspec.name == 'private':
            self._private = val
        else:
            raise ValueError('Unknown property %r' % pspec.name)

        self._publish_properties()

    def set_private(self, val, reply_handler, error_handler):
        _logger.debug('set_private %r' % val)
        self._activity.SetProperties({'private': bool(val)},
                                     reply_handler=reply_handler,
                                     error_handler=error_handler)

    def get_joined_buddies(self):
        """Retrieve the set of Buddy objects attached to this activity

        returns list of presence Buddy objects that we can successfully
        create from the buddy object paths that PS has for this activity.
        """
        return list(self._joined_buddies.values())

    def get_buddy_by_handle(self, handle):
        """Retrieve the Buddy object given a telepathy handle.

        buddy object paths are cached in self._handle_to_buddy_path,
        so we can get the buddy without calling PS.
        """
        object_path = self._handle_to_buddy_path.get(handle, None)
        if object_path:
            buddy = self._ps_new_object(object_path)
            return buddy
        return None

    def invite(self, buddy, message, response_cb):
        """Invite the given buddy to join this activity.

        The callback will be called with one parameter: None on success,
        or an exception on failure.
        """
        if not self._joined:
            raise RuntimeError('Cannot invite a buddy to an activity that is'
                               'not shared.')
        self.telepathy_text_chan[CHANNEL].AddMembers(
            [buddy.contact_handle], message,
            dbus_interface=CHANNEL_INTERFACE_GROUP,
            reply_handler=partial(
                self.__invite_cb, response_cb),
            error_handler=partial(self.__invite_cb, response_cb))

    def __invite_cb(self, response_cb, error=None):
        response_cb(error)

    def set_up_tubes(self, reply_handler, error_handler):
        raise NotImplementedError()

    def __joined_cb(self, join_command, error):
        _logger.debug('%r: Join finished %r' % (self, error))
        if error is not None:
            self.emit('joined', error is None, str(error))
        self.telepathy_text_chan = join_command.text_channel
        self.telepathy_tubes_chan = join_command.tubes_channel
        self._channel_self_handle = join_command.channel_self_handle
        self._text_channel_group_flags = join_command.text_channel_group_flags
        self._start_tracking_buddies()
        self._start_tracking_channel()

    def _start_tracking_buddies(self):
        group = self.telepathy_text_chan[CHANNEL_INTERFACE_GROUP]

        group.GetAllMembers(reply_handler=self.__get_all_members_cb,
                            error_handler=self.__error_handler_cb)

        group.connect_to_signal('MembersChanged',
                                self.__text_channel_members_changed_cb)

    def _start_tracking_channel(self):
        channel = self.telepathy_text_chan[CHANNEL]
        channel.connect_to_signal('Closed', self.__text_channel_closed_cb)

    def __get_all_members_cb(self, members, local_pending, remote_pending):
        _logger.debug(
            '__get_all_members_cb %r %r' %
            (members, self._text_channel_group_flags))
        if self._channel_self_handle in members:
            members.remove(self._channel_self_handle)

        if not members:
            return

        self._resolve_handles(members, reply_cb=self._add_initial_buddies)

    def _resolve_handles(self, input_handles, reply_cb):
        def get_handle_owners_cb(handles):
            self.telepathy_conn.InspectHandles(
                HANDLE_TYPE_CONTACT, handles,
                reply_handler=reply_cb,
                error_handler=self.__error_handler_cb,
                dbus_interface=CONNECTION)

        if self._text_channel_group_flags & \
                CHANNEL_GROUP_FLAG_CHANNEL_SPECIFIC_HANDLES:

            group = self.telepathy_text_chan[CHANNEL_INTERFACE_GROUP]
            group.GetHandleOwners(input_handles,
                                  reply_handler=get_handle_owners_cb,
                                  error_handler=self.__error_handler_cb)
        else:
            get_handle_owners_cb(input_handles)

    def _add_initial_buddies(self, contact_ids):
        _logger.debug('__add_initial_buddies %r' % contact_ids)
        for contact_id in contact_ids:
            self._buddies[contact_id] = self._get_buddy(contact_id)
            self._joined_buddies[contact_id] = self._get_buddy(contact_id)
        # Once we have the initial members, we can finish the join process
        self._joined = True
        self.emit('joined', True, None)

    def __text_channel_members_changed_cb(self, message, added, removed,
                                          local_pending, remote_pending,
                                          actor, reason):
        _logger.debug('__text_channel_members_changed_cb %r' %
                      [added, message, added, removed, local_pending,
                       remote_pending, actor, reason])
        if self._channel_self_handle in added:
            added.remove(self._channel_self_handle)
        if added:
            self._resolve_handles(added, reply_cb=self._add_buddies)

        if self._channel_self_handle in removed:
            removed.remove(self._channel_self_handle)
        if removed:
            self._resolve_handles(removed, reply_cb=self._remove_buddies)

    def _add_buddies(self, contact_ids):
        for contact_id in contact_ids:
            if contact_id not in self._buddies:
                buddy = self._get_buddy(contact_id)
                self.emit('buddy-joined', buddy)
                self._buddies[contact_id] = buddy
            if contact_id not in self._joined_buddies:
                self._joined_buddies[contact_id] = buddy

    def _remove_buddies(self, contact_ids):
        for contact_id in contact_ids:
            if contact_id in self._buddies:
                buddy = self._get_buddy(contact_id)
                self.emit('buddy-left', buddy)
                del self._buddies[contact_id]

    def _get_buddy(self, contact_id):
        if contact_id in self._buddies:
            return self._buddies[contact_id]
        else:
            return Buddy(self._account_path, contact_id)

    def join(self):
        """Join this activity.

        Emits 'joined' and otherwise does nothing if we're already joined.
        """
        if self._join_command is not None:
            return

        if self._joined:
            self.emit('joined', True, None)
            return

        _logger.debug('%r: joining' % self)

        self._join_command = _JoinCommand(self.telepathy_conn,
                                          self.room_handle)
        self._join_command.connect('finished', self.__joined_cb)
        self._join_command.run()

    def share(self, share_activity_cb, share_activity_error_cb):
        if self.room_handle is not None:
            raise ValueError('Already have a room handle')

        self._share_command = _ShareCommand(self.telepathy_conn, self._id)
        self._share_command.connect('finished',
                                    partial(self.__shared_cb,
                                            share_activity_cb,
                                            share_activity_error_cb))
        self._share_command.run()

    def __shared_cb(self, share_activity_cb, share_activity_error_cb,
                    share_command, error):
        _logger.debug('%r: Share finished %r' % (self, error))
        if error is None:
            self._joined = True
            self.room_handle = share_command.room_handle
            self.telepathy_text_chan = share_command.text_channel
            self.telepathy_tubes_chan = share_command.tubes_channel
            self._channel_self_handle = share_command.channel_self_handle
            self._text_channel_group_flags = \
                share_command.text_channel_group_flags
            self._publish_properties()
            self._start_tracking_properties()
            self._start_tracking_buddies()
            self._start_tracking_channel()
            share_activity_cb(self)
        else:
            share_activity_error_cb(self, error)

    def _publish_properties(self):
        properties = {}

        if self._color is not None:
            properties['color'] = str(self._color)
        if self._name is not None:
            properties['name'] = str(self._name)
        if self._type is not None:
            properties['type'] = self._type
        if self._tags is not None:
            properties['tags'] = self._tags
        properties['private'] = self._private

        self.telepathy_conn.SetProperties(
            self.room_handle,
            properties,
            dbus_interface=CONN_INTERFACE_ACTIVITY_PROPERTIES)

    def __share_error_cb(self, share_activity_error_cb, error):
        logging.debug('%r: Share failed because: %s' % (self, error))
        share_activity_error_cb(self, error)

    # GetChannels() wrapper

    def get_channels(self):
        """Retrieve communications channel descriptions for the activity

        Returns a tuple containing:
            - the D-Bus well-known service name of the connection
              (FIXME: this is redundant; in Telepathy it can be derived
              from that of the connection)
            - the D-Bus object path of the connection
            - a list of D-Bus object paths representing the channels
              associated with this activity
        """
        bus_name = self.telepathy_conn.requested_bus_name
        connection_path = self.telepathy_conn.object_path
        channels = [self.telepathy_text_chan.object_path,
                    self.telepathy_tubes_chan.object_path]

        _logger.debug('%r: bus name is %s, connection is %s, channels are %r' %
                      (self, bus_name, connection_path, channels))
        return bus_name, connection_path, channels

    # Leaving
    def __text_channel_closed_cb(self):
        self._joined = False
        self.emit('joined', False, 'left activity')

    def leave(self):
        """Leave this shared activity"""
        _logger.debug('%r: leaving' % self)
        self.telepathy_text_chan[CHANNEL].Close()


class _BaseCommand(GObject.GObject):
    __gsignals__ = {
        'finished': (GObject.SignalFlags.RUN_FIRST, None,
                     ([object])),
    }

    def __init__(self):
        GObject.GObject.__init__(self)

        self.text_channel = None
        self.text_channel_group_flags = None
        self.tubes_channel = None
        self.room_handle = None
        self.channel_self_handle = None

    def run(self):
        raise NotImplementedError()


class _ShareCommand(_BaseCommand):
    def __init__(self, connection, activity_id):
        _BaseCommand.__init__(self)

        self._connection = connection
        self._activity_id = activity_id
        self._finished = False
        self._join_command = None

    def run(self):
        self._connection.RequestHandles(
            HANDLE_TYPE_ROOM,
            [self._activity_id],
            reply_handler=self.__got_handles_cb,
            error_handler=self.__error_handler_cb,
            dbus_interface=CONNECTION)

    def __got_handles_cb(self, handles):
        logging.debug('__got_handles_cb %r' % handles)
        self.room_handle = handles[0]

        self._join_command = _JoinCommand(self._connection, self.room_handle)
        self._join_command.connect('finished', self.__joined_cb)
        self._join_command.run()

    def __joined_cb(self, join_command, error):
        _logger.debug('%r: Join finished %r' % (self, error))
        if error is not None:
            self._finished = True
            self.emit('finished', error)
            return

        self.text_channel = join_command.text_channel
        self.text_channel_group_flags = join_command.text_channel_group_flags
        self.tubes_channel = join_command.tubes_channel
        self.channel_self_handle = join_command.channel_self_handle

        self._connection.AddActivity(
            self._activity_id,
            self.room_handle,
            reply_handler=self.__added_activity_cb,
            error_handler=self.__error_handler_cb,
            dbus_interface=CONN_INTERFACE_BUDDY_INFO)

    def __added_activity_cb(self):
        self._finished = True
        self.emit('finished', None)

    def __error_handler_cb(self, error):
        self._finished = True
        self.emit('finished', error)


class _JoinCommand(_BaseCommand):
    def __init__(self, connection, room_handle):
        _BaseCommand.__init__(self)

        self._connection = connection
        self._finished = False
        self.room_handle = room_handle
        self._global_self_handle = None
        self._tubes_supported = True

    def run(self):
        if self._finished:
            raise RuntimeError('This command has already finished')

        self._connection.Get(CONNECTION, 'SelfHandle',
                             reply_handler=self.__get_self_handle_cb,
                             error_handler=self.__error_handler_cb,
                             dbus_interface=PROPERTIES_IFACE)

    def __get_self_handle_cb(self, handle):
        self._global_self_handle = handle

        self._text_ready = False
        self._connection.RequestChannel(
            CHANNEL_TYPE_TEXT,
            HANDLE_TYPE_ROOM,
            self.room_handle, True,
            reply_handler=self.__create_text_channel_cb,
            error_handler=self.__error_handler_cb,
            dbus_interface=CONNECTION)

        self._tubes_ready = False
        self._connection.RequestChannel(
            CHANNEL_TYPE_TUBES,
            HANDLE_TYPE_ROOM,
            self.room_handle,
            True,
            reply_handler=self.__create_tubes_channel_cb,
            error_handler=self.__tubes_error_handler_cb,
            dbus_interface=CONNECTION)

    def __create_text_channel_cb(self, channel_path):
        self.text_channel = {}
        self.text_proxy = dbus.Bus().get_object(
            self._connection.requested_bus_name, channel_path)
        self.text_channel[PROPERTIES_IFACE] = dbus.Interface(
            self.text_proxy, PROPERTIES_IFACE)
        self.text_channel[CHANNEL_TYPE_TEXT] = \
            dbus.Interface(self.text_proxy, CHANNEL_TYPE_TEXT)
        self.text_channel[CHANNEL] = dbus.Interface(self.text_proxy, CHANNEL)
        self.text_channel[CHANNEL].GetInterfaces(
            reply_handler=self.__text_get_interfaces_reply_cb,
            error_handler=self.__error_handler_cb)

    def __text_get_interfaces_reply_cb(self, interfaces):
        for interface in interfaces:
            self.text_channel[interface] = dbus.Interface(
                self.text_proxy, interface)
        _logger.debug('%r: Text channel is ready' % (self))
        self._text_ready = True
        self._ready()

    def __create_tubes_channel_cb(self, channel_path):
        self.tubes_channel = {}
        self.tubes_proxy = dbus.Bus().get_object(
            self._connection.requested_bus_name, channel_path)
        self.tubes_channel[PROPERTIES_IFACE] = dbus.Interface(
            self.tubes_proxy, PROPERTIES_IFACE)
        self.tubes_channel[CHANNEL_TYPE_TUBES] = \
            dbus.Interface(self.tubes_proxy, CHANNEL_TYPE_TUBES)
        self.tubes_channel[CHANNEL] = dbus.Interface(self.tubes_proxy, CHANNEL)
        self.tubes_channel[CHANNEL].GetInterfaces(
            reply_handler=self.__tubes_get_interfaces_reply_cb,
            error_handler=self.__error_handler_cb)

    def __tubes_get_interfaces_reply_cb(self, interfaces):
        for interface in interfaces:
            self.tubes_channel[interface] = dbus.Interface(
                self.tubes_proxy, interface)
        _logger.debug('%r: Tubes channel is ready' % (self))
        self._tubes_ready = True
        self._ready()

    def __tubes_error_handler_cb(self, error):
        if (error.get_dbus_name() ==
                'org.freedesktop.Telepathy.Error.NotImplemented'):
            self._tubes_supported = False
            self._ready()
        else:
            self._finished = True
            self.emit('finished', error)

    def __error_handler_cb(self, error):
        self._finished = True
        self.emit('finished', error)

    def _ready(self):
        if not self._text_ready or \
                (self._tubes_supported and not self._tubes_ready):
            return

        _logger.debug('%r: finished setting up channel' % self)

        self._add_self_to_channel()

    def __text_channel_group_flags_changed_cb(self, added, removed):
        _logger.debug(
            '__text_channel_group_flags_changed_cb %r %r' %
            (added, removed))
        self.text_channel_group_flags |= added
        self.text_channel_group_flags &= ~removed

    def _add_self_to_channel(self):
        # FIXME: cope with non-Group channels here if we want to support
        # non-OLPC-compatible IMs

        group = self.text_channel[CHANNEL_INTERFACE_GROUP]

        def got_all_members(members, local_pending, remote_pending):
            _logger.debug('got_all_members members %r local_pending %r '
                          'remote_pending %r' % (members, local_pending,
                                                 remote_pending))

            if self.text_channel_group_flags & \
                    CHANNEL_GROUP_FLAG_CHANNEL_SPECIFIC_HANDLES:
                self_handle = self.channel_self_handle
            else:
                self_handle = self._global_self_handle

            if self_handle in local_pending:
                _logger.debug('%r: We are in local pending - entering' % self)
                group.AddMembers([self_handle], '',
                                 reply_handler=lambda: None,
                                 error_handler=lambda e: self._join_failed_cb(
                                     e, 'got_all_members AddMembers'))

            if members:
                self.__text_channel_members_changed_cb('', members, (),
                                                       (), (), 0, 0)

        def got_group_flags(flags):
            self.text_channel_group_flags = flags
            # by the time we hook this, we need to know the group flags
            group.connect_to_signal('MembersChanged',
                                    self.__text_channel_members_changed_cb)

            # bootstrap by getting the current state. This is where we find
            # out whether anyone was lying to us in their PEP info
            group.GetAllMembers(reply_handler=got_all_members,
                                error_handler=self.__error_handler_cb)

        def got_self_handle(channel_self_handle):
            self.channel_self_handle = channel_self_handle
            group.connect_to_signal('GroupFlagsChanged',
                                    self.__text_channel_group_flags_changed_cb)
            group.GetGroupFlags(reply_handler=got_group_flags,
                                error_handler=self.__error_handler_cb)

        group.GetSelfHandle(reply_handler=got_self_handle,
                            error_handler=self.__error_handler_cb)

    def __text_channel_members_changed_cb(self, message, added, removed,
                                          local_pending, remote_pending,
                                          actor, reason):
        _logger.debug('__text_channel_members_changed_cb added %r removed %r '
                      'local_pending %r remote_pending %r channel_self_handle '
                      '%r' % (added, removed, local_pending, remote_pending,
                              self.channel_self_handle))

        if self.text_channel_group_flags & \
                CHANNEL_GROUP_FLAG_CHANNEL_SPECIFIC_HANDLES:
            self_handle = self.channel_self_handle
        else:
            self_handle = self._global_self_handle

        if self_handle not in added:
            return

        # Use RoomConfig1 to configure the text channel. If this
        # doesn't exist, fall-back on old-style PROPERTIES_INTERFACE.
        if CONN_INTERFACE_ROOM_CONFIG in self.text_channel:
            self.__update_room_config()
        elif PROPERTIES_INTERFACE in self.text_channel:
            self.text_channel[PROPERTIES_INTERFACE].ListProperties(
                reply_handler=self.__list_properties_cb,
                error_handler=self.__error_handler_cb)
        else:
            # FIXME: when does this codepath get hit?
            # It could be related to no property configuration being available
            # in the selected backend, or it could be called at some stage
            # of the protocol when properties aren't available yet.
            self._finished = True
            self.emit('finished', None)

    def __update_room_config(self):
        # FIXME: invite-only ought to be set on private activities; but
        # since only the owner can change invite-only, that would break
        # activity scope changes.
        props = {
            # otherwise buddy resolution breaks
            'Anonymous': False,
            # anyone who knows about the channel can join
            'InviteOnly': False,
            # vanish when there are no members
            'Persistent': False,
            # don't appear in server room lists
            'Private': True,
        }
        room_cfg = self.text_channel[CONN_INTERFACE_ROOM_CONFIG]
        room_cfg.UpdateConfiguration(props,
                                     reply_handler=self.__room_cfg_updated_cb,
                                     error_handler=self.__room_cfg_error_cb)

    def __room_cfg_updated_cb(self):
        self._finished = True
        self.emit('finished', None)

    def __room_cfg_error_cb(self, error):
        # If RoomConfig update fails, it's probably because we don't have
        # permission (e.g. we are not the session initiator). Thats OK -
        # ignore the failure and carry on.
        if (error.get_dbus_name() !=
                'org.freedesktop.Telepathy.Error.PermissionDenied'):
            logging.error("Error setting room configuration: %s", error)
        self._finished = True
        self.emit('finished', None)

    def __list_properties_cb(self, prop_specs):
        # FIXME: invite-only ought to be set on private activities; but
        # since only the owner can change invite-only, that would break
        # activity scope changes.
        props = {
            # otherwise buddy resolution breaks
            'anonymous': False,
            # anyone who knows about the channel can join
            'invite-only': False,
            # so non-owners can invite others
            'invite-restricted': False,
            # vanish when there are no members
            'persistent': False,
            # don't appear in server room lists
            'private': True,
        }
        props_to_set = []
        for ident, name, sig_, flags in prop_specs:
            value = props.pop(name, None)
            if value is not None:
                if flags & PROPERTY_FLAG_WRITE:
                    props_to_set.append((ident, value))
                # FIXME: else error, but only if we're creating the room?
        # FIXME: if props is nonempty, then we want to set props that aren't
        # supported here - raise an error?

        if props_to_set:
            self.text_channel[PROPERTIES_INTERFACE].SetProperties(
                props_to_set, reply_handler=self.__set_properties_cb,
                error_handler=self.__error_handler_cb)
        else:
            self._finished = True
            self.emit('finished', None)

    def __set_properties_cb(self):
        self._finished = True
        self.emit('finished', None)



# Copyright (C) 2014, Sugarlabs
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""
The power module provides an interface to *powerd*, a daemon that
manages the aggressive suspend and wakeup policies for early OLPC
laptops.

The module does nothing if *powerd* is not present.  *powerd* is not
required on laptops other than OLPC XO-1, XO-1.5, XO-1.75 and XO-4.
Distributions of Sugar on other hardware need not include the *powerd*
package.
"""

import os
import logging

_POWERD_INHIBIT_DIR = '/var/run/powerd-inhibit-suspend'

_power_manager = None


def get_power_manager():
    """
    Get the power manager instance.  Only one instance exists, and
    will always be returned.

    Returns:
        an instance of :class:`sugar3.power.PowerManager`.
    """
    global _power_manager
    if _power_manager is None:
        _power_manager = PowerManager()
    return _power_manager


class PowerManager():
    """
    Control of automatic idle suspend, with reference counting.

    :class:`sugar3.activity.activity.Activity` calls
    :py:meth:`inhibit_suspend` before speaking text, or when an
    activity collaboration begins.

    Activities may call :py:meth:`inhibit_suspend` before playing
    music, video, speaking large amounts of text, collaborating, or
    waiting for response to network operations.

    As an example, the Clock activity inhibits automatic idle suspend
    while it is active, so that the displayed clock-face continues to
    change.  Otherwise it would freeze.

    :class:`sugar3.activity.activity.Activity` calls
    :py:meth:`shutdown` as an activity terminates, in case the
    activity has failed to call :py:meth:`restore_suspend`.

    While automatic idle suspend is inhibited, *powerd* will
    continue to dim and blank the display.

    Both the :py:meth:`inhibit_suspend` and :py:meth:`restore_suspend`
    methods are reference counted; automatic idle suspend is not
    restored until the same number of calls to restore are made.

    *powerd* is resilient against failure to restore automatic idle
    suspend; it verifies an inhibit request and deletes it if the
    requesting process has terminated.
    """

    def __init__(self):
        self._suspend_inhibit_counter = 0
        if os.path.exists(_POWERD_INHIBIT_DIR):
            self._path = os.path.join(_POWERD_INHIBIT_DIR, str(os.getpid()))
        else:
            self._path = None

    def __del__(self):
        self._remove_flag_file()

    def suspend_breaks_collaboration(self):
        """
        Does automatic idle suspend break collaboration with this
        toolkit?  Yes.  For future use by a toolkit with more
        resilient collaboration.

        Returns:
            True
        """
        return True

    def inhibit_suspend(self):
        """
        Inhibit automatic idle suspend until restored.
        """
        if self._path and self._suspend_inhibit_counter == 0:
            try:
                with open(self._path, 'w') as flag_file:
                    flag_file.write('')
            except IOError:
                logging.error("Inhibit Suspend: Could not create file %s",
                              self._path)

        self._suspend_inhibit_counter += 1

    def restore_suspend(self):
        """
        Possibly restore automatic idle suspend.
        """
        self._suspend_inhibit_counter -= 1
        if self._suspend_inhibit_counter > 0:
            return
        self._remove_flag_file()

    def is_suspend_inhibited(self):
        """
        Check if automatic idle suspend is inhibited.

        Returns:
            inhibited (bool): whether automatic idle suspend is inhibited.
        """
        return self._suspend_inhibit_counter > 0

    def shutdown(self):
        """
        Shutdown the power manager.

        Restores automatic idle suspend regardless of reference counting.
        """
        self._remove_flag_file()

    def _remove_flag_file(self):
        if self._path:
            try:
                os.unlink(self._path)
            except OSError:
                pass
        self._suspend_inhibit_counter = 0



# Copyright (C) 2006-2007 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

'''
Icons are small pictures that are used to decorate components.  In Sugar, icons
are SVG files that are re-coloured with a fill and a stroke colour.  Typically,
icons representing the system use a greyscale color palette, whereas icons
representing people take on their selected XoColors.

Designing a Sugar Icon
======================

If you want to make an icon to use in Sugar, start by designing something in
a vector program, like Inkscape.  When you are designing the icon, use a canvas
that is 55x55px.

When designing your icon, use 2 colors.  For example, use a black stroke and a
white fill color.  You need to keep this consistent and remember those colors.

You should also keep the subcell grid in mind when designing the icon.  A grid
cell (which the size of an icon) is made up of 5 by 5 subcells.  To use this in
Inkscape, enable the page grid (View -> Page Grid), the go to grid properties
(File -> Document Properties -> Grids).  In grid properties, set the "Spacing
X" option to 11, "Spacing Y" to 11 and "Major grid line every" to 1.

Before your icon is ready to be used in Sugar, it needs to be Sugarized.  This
converts the colors to SVG entities, which allows Sugar to change the colors
of the icon.  To do that, just run the `sugar-iconify`__ script.  Usually, it
"just works" like::

    python path/to/sugar-iconify.py -o icon.svg

__ https://github.com/GhostAlgorithm/sugariconify/blob/master/sugariconify.py

Code Example
============

Example of using icons with badges:

.. literalinclude:: ../examples/iconbadges.py

Badge Icons
===========
Badge icons are small icons that are attached to the normal icon.
For example, a WiFi network icon might have a star badge attached to
the bottom left corner (the "attach point") that indicates that the
network is connected to.

Badge icons are displayed at _BADGE_SIZE percent (45% currently),
of their main icon.

Badge icons are specified by the icon name, in the same sense that
normal icons have a :any:`Icon.set_icon_name` function.

Attach Points
-------------

Where the badge icon is placed is defined by the main icon.  By
default, it is centered on 0, 0.  That means that the 3 quarters of
the icon will be cut off!  Therefore, it is helpful to define the
attach points.

When Sugar loads the main icon, it looks for a `.icon` file.  For
example, if the icon path is resolved to `/theme/computer-xo.svg`,
the `/theme/computer-xo.icon` will be tried to find the attach points.

The `.icon` files are to be formatted as follows::

    [Icon Data]
    AttachPoints=970,850

In this example, the badge will be centered at 97.0% on the X axis,
and 85.0% on the Y axis.
'''

import six
import re
import math
import logging
import os

from six.moves.configparser import ConfigParser

import gi
gi.require_version('Rsvg', '2.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Rsvg
import cairo

from sugar3.graphics import style
from sugar3.graphics.xocolor import XoColor
from sugar3.util import LRU

_BADGE_SIZE = 0.45


class _SVGLoader(object):

    def __init__(self):
        self._cache = LRU(100)

    def load(self, file_name, entities, cache):
        if file_name in self._cache:
            icon = self._cache[file_name]
        else:
            icon_file = open(file_name, 'r')
            icon = icon_file.read()
            icon_file.close()

            if cache:
                self._cache[file_name] = icon

        for entity, value in list(entities.items()):
            if isinstance(value, six.string_types):
                xml = '<!ENTITY %s "%s">' % (entity, value)
                icon = re.sub('<!ENTITY %s .*>' % entity, xml, icon)
            else:
                logging.error(
                    'Icon %s, entity %s is invalid.', file_name, entity)

        return Rsvg.Handle.new_from_data(icon.encode('utf-8'))


class _IconInfo(object):

    def __init__(self):
        self.file_name = None
        self.attach_x = 0
        self.attach_y = 0


class _BadgeInfo(object):

    def __init__(self):
        self.attach_x = 0
        self.attach_y = 0
        self.size = 0
        self.icon_padding = 0


class _IconBuffer(object):

    _surface_cache = LRU(100)
    _loader = _SVGLoader()

    def __init__(self):
        self.icon_name = None
        self.icon_size = None
        self.file_name = None
        self.fill_color = None
        self.background_color = None
        self.stroke_color = None
        self.badge_name = None
        self.width = None
        self.height = None
        self.cache = False
        self.scale = 1.0
        self.pixbuf = None

    def _get_cache_key(self, sensitive):
        if self.background_color is None:
            color = None
        else:
            color = (self.background_color.red, self.background_color.green,
                     self.background_color.blue)

        return (self.icon_name, self.file_name, self.pixbuf, self.fill_color,
                self.stroke_color, self.badge_name, self.width, self.height,
                color, sensitive)

    def _load_svg(self, file_name):
        entities = {}
        if self.fill_color:
            entities['fill_color'] = self.fill_color
        if self.stroke_color:
            entities['stroke_color'] = self.stroke_color

        return self._loader.load(file_name, entities, self.cache)

    def _get_attach_points(self, info, size_request):
        has_attach_points_, attach_points = info.get_attach_points()
        attach_x = attach_y = 0
        if attach_points:
            # this works only for Gtk < 3.14
            # https://developer.gnome.org/gtk3/stable/GtkIconTheme.html
            # #gtk-icon-info-get-attach-points
            attach_x = float(attach_points[0].x) / size_request
            attach_y = float(attach_points[0].y) / size_request
        elif info.get_filename():
            # try read from the .icon file
            icon_filename = info.get_filename().replace('.svg', '.icon')
            if icon_filename != info.get_filename() and \
                    os.path.exists(icon_filename):

                try:
                    with open(icon_filename) as config_file:
                        cp = ConfigParser()
                        cp.readfp(config_file)
                        attach_points_str = cp.get('Icon Data', 'AttachPoints')
                        attach_points = attach_points_str.split(',')
                        attach_x = float(attach_points[0].strip()) / 1000
                        attach_y = float(attach_points[1].strip()) / 1000
                except Exception as e:
                    logging.exception('Exception reading icon info: %s', e)

        return attach_x, attach_y

    def _get_icon_info(self, file_name, icon_name):
        icon_info = _IconInfo()

        if file_name:
            icon_info.file_name = file_name
        elif icon_name:
            theme = Gtk.IconTheme.get_default()

            size = 50
            if self.width is not None:
                size = self.width

            info = theme.lookup_icon(icon_name, int(size), 0)
            if info:
                attach_x, attach_y = self._get_attach_points(info, size)

                icon_info.file_name = info.get_filename()
                icon_info.attach_x = attach_x
                icon_info.attach_y = attach_y

                del info
            else:
                logging.warning('No icon with the name %s was found in the '
                                'theme.', icon_name)

        return icon_info

    def _draw_badge(self, context, size, sensitive, widget):
        theme = Gtk.IconTheme.get_default()
        badge_info = theme.lookup_icon(self.badge_name, int(size), 0)
        if badge_info:
            badge_file_name = badge_info.get_filename()
            if badge_file_name.endswith('.svg'):
                handle = self._loader.load(badge_file_name, {}, self.cache)

                icon_width = handle.props.width
                icon_height = handle.props.height

                pixbuf = handle.get_pixbuf()
            else:
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(badge_file_name)

                icon_width = pixbuf.get_width()
                icon_height = pixbuf.get_height()

            context.scale(float(size) / icon_width,
                          float(size) / icon_height)

            if not sensitive:
                pixbuf = self._get_insensitive_pixbuf(pixbuf, widget)
            Gdk.cairo_set_source_pixbuf(context, pixbuf, 0, 0)
            context.paint()

    def _get_size(self, icon_width, icon_height, padding):
        if self.width is not None and self.height is not None:
            width = self.width + padding
            height = self.height + padding
        else:
            width = icon_width + padding
            height = icon_height + padding

        return width, height

    def _get_badge_info(self, icon_info, icon_width, icon_height):
        info = _BadgeInfo()
        if self.badge_name is None:
            return info

        info.size = int(_BADGE_SIZE * icon_width)
        info.attach_x = int(icon_info.attach_x * icon_width - info.size / 2)
        info.attach_y = int(icon_info.attach_y * icon_height - info.size / 2)

        if info.attach_x < 0 or info.attach_y < 0:
            info.icon_padding = max(-info.attach_x, -info.attach_y)
        elif info.attach_x + info.size > icon_width or \
                info.attach_y + info.size > icon_height:
            x_padding = info.attach_x + info.size - icon_width
            y_padding = info.attach_y + info.size - icon_height
            info.icon_padding = max(x_padding, y_padding)

        return info

    def _get_xo_color(self):
        if self.stroke_color and self.fill_color:
            return XoColor('%s,%s' % (self.stroke_color, self.fill_color))
        else:
            return None

    def _set_xo_color(self, xo_color):
        if xo_color:
            self.stroke_color = xo_color.get_stroke_color()
            self.fill_color = xo_color.get_fill_color()
        else:
            self.stroke_color = None
            self.fill_color = None

    def _get_insensitive_pixbuf(self, pixbuf, widget):
        if not (widget and widget.get_style()):
            return pixbuf

        icon_source = Gtk.IconSource()
        # Special size meaning "don't touch"
        icon_source.set_size(-1)
        icon_source.set_pixbuf(pixbuf)
        icon_source.set_state(Gtk.StateType.INSENSITIVE)
        icon_source.set_direction_wildcarded(False)
        icon_source.set_size_wildcarded(False)

        widget_style = widget.get_style()
        pixbuf = widget_style.render_icon(
            icon_source, widget.get_direction(),
            Gtk.StateType.INSENSITIVE, -1, widget,
            'sugar-icon')

        return pixbuf

    def get_surface(self, sensitive=True, widget=None):
        cache_key = self._get_cache_key(sensitive)
        if cache_key in self._surface_cache:
            return self._surface_cache[cache_key]

        if self.pixbuf:
            # We alredy have the pixbuf for this icon.
            pixbuf = self.pixbuf
            icon_width = pixbuf.get_width()
            icon_height = pixbuf.get_height()
            icon_info = self._get_icon_info(self.file_name, self.icon_name)
            is_svg = False
        else:
            # We run two attempts at finding the icon. First, we try the icon
            # requested by the user. If that fails, we fall back on
            # document-generic. If that doesn't work out, bail.
            icon_width = None
            for (file_name, icon_name) in ((self.file_name, self.icon_name),
                                           (None, 'document-generic')):
                icon_info = self._get_icon_info(file_name, icon_name)
                if icon_info.file_name is None:
                    return None

                is_svg = icon_info.file_name.endswith('.svg')

                if is_svg:
                    try:
                        handle = self._load_svg(icon_info.file_name)
                        icon_width = handle.props.width
                        icon_height = handle.props.height
                        break
                    except IOError:
                        pass
                else:
                    try:
                        path = icon_info.file_name
                        pixbuf = GdkPixbuf.Pixbuf.new_from_file(path)
                        icon_width = pixbuf.get_width()
                        icon_height = pixbuf.get_height()
                        break
                    except GLib.GError:
                        pass

        if icon_width is None:
            # Neither attempt found an icon for us to use
            return None

        badge_info = self._get_badge_info(icon_info, icon_width, icon_height)

        padding = badge_info.icon_padding
        width, height = self._get_size(icon_width, icon_height, padding)
        if self.background_color is None:
            surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, int(width),
                                         int(height))
            context = cairo.Context(surface)
        else:
            surface = cairo.ImageSurface(cairo.FORMAT_RGB24, int(width),
                                         int(height))
            context = cairo.Context(surface)
            context.set_source_color(self.background_color)
            context.paint()

        context.scale(float(width) / (icon_width + padding * 2),
                      float(height) / (icon_height + padding * 2))
        context.save()

        context.translate(padding, padding)
        if is_svg:
            if sensitive:
                handle.render_cairo(context)
            else:
                pixbuf = handle.get_pixbuf()
                pixbuf = self._get_insensitive_pixbuf(pixbuf, widget)

                Gdk.cairo_set_source_pixbuf(context, pixbuf, 0, 0)
                context.paint()
        else:
            if not sensitive:
                pixbuf = self._get_insensitive_pixbuf(pixbuf, widget)
            Gdk.cairo_set_source_pixbuf(context, pixbuf, 0, 0)
            context.paint()

        if self.badge_name:
            context.restore()
            context.translate(badge_info.attach_x, badge_info.attach_y)
            self._draw_badge(context, badge_info.size, sensitive, widget)

        self._surface_cache[cache_key] = surface

        return surface

    xo_color = property(_get_xo_color, _set_xo_color)


class Icon(Gtk.Image):
    '''
    The most basic Sugar icon class.  Displays the icon given.

    You must set either the `file_name`, `file` or `icon_name` properties,
    otherwise, no icon will be visible.

    You should set the `pixel_size`, using constants the `*_ICON_SIZE`
    constants from :any:`sugar3.graphics.style`.

    You should set the color (either via `xo_color` or `fill_color` and
    `stroke_color`), otherwise the default black and white fill and stroke
    will be used.

    Keyword Args:
        file_name (str): a path to the SVG icon file
        file (object): same behaviour as file_name, but for
            :class:`sugar3.util.TempFilePath` type objects
        icon_name (str): a name of an icon in the theme to display.  The
            icons in the theme include those in the sugar-artwork project
            and icons in the activity's '/icons' directory
        pixel_size (int): size of the icon, in pixels.  Best to use the
            constants from :class:`sugar3.graphics.style`, as those constants
            are scaled based on the user's preferences
        xo_color (sugar3.graphics.xocolor.XoColor): color to display icon,
            a shortcut that just sets the fill_color and stroke_color
        fill_color (str): a string, like '#FFFFFF', that will serve as the
            fill color for the icon
        stroke_color (str): a string, like '#282828', that will serve as the
            stroke color for the icon
        icon_size: deprecated since 0.102.0, use pixel_size instead
        badge_name (str): the icon_name for a badge icon,
            see :any:`set_badge_name`
        alpha (float): transparency of the icon, defaults to 1.0
    '''

    __gtype_name__ = 'SugarIcon'

    _MENU_SIZES = (Gtk.IconSize.MENU, Gtk.IconSize.DND,
                   Gtk.IconSize.SMALL_TOOLBAR, Gtk.IconSize.BUTTON)

    def __init__(self, **kwargs):
        self._buffer = _IconBuffer()
        # HACK: need to keep a reference to the path so it doesn't get garbage
        # collected while it's still used if it's a sugar3.util.TempFilePath.
        # See #1175
        self._file = None
        self._alpha = 1.0
        self._scale = 1.0

        if 'icon_size' in kwargs:
            logging.warning("icon_size is deprecated. Use pixel_size instead.")

        GObject.GObject.__init__(self, **kwargs)

    def get_file(self):
        return self._file

    def set_file(self, file_name):
        self._file = file_name
        self._buffer.file_name = file_name

    file = GObject.Property(type=object, setter=set_file, getter=get_file)

    def get_pixbuf(self):
        '''
        Returns the :class:`GdkPixbuf.Pixbuf` for the icon, if one has been
        loaded yet.  If the icon has been drawn (:any:`do_draw`), the icon
        will be loaded.

        The pixbuf only contains the SVG icon that has been loaded and
        recoloured.  It does not contain the badge.
        '''
        return self._buffer.pixbuf

    def set_pixbuf(self, pixbuf):
        '''
        Set the pixbuf.  This will force the icon to be rendered with the
        given pixbuf.  The icon will still be centered, badge added, etc.

        Args:
            pixbuf (GdkPixbuf.Pixbuf): pixbuf to set
        '''
        self._buffer.pixbuf = pixbuf

    pixbuf = GObject.Property(type=object, setter=set_pixbuf,
                              getter=get_pixbuf)
    '''
    icon.props.pixbuf -> see :any:`get_pixbuf` and :any:`set_pixbuf`
    '''

    def _sync_image_properties(self):
        if self._buffer.icon_name != self.props.icon_name:
            self._buffer.icon_name = self.props.icon_name

        if self._buffer.file_name != self.props.file:
            self._buffer.file_name = self.props.file

        pixel_size = None
        if self.props.pixel_size == -1:
            if self.props.icon_size in self._MENU_SIZES:
                pixel_size = style.SMALL_ICON_SIZE
            else:
                pixel_size = style.STANDARD_ICON_SIZE
        else:
            pixel_size = self.props.pixel_size

        width = height = pixel_size

        if self._buffer.width != width or self._buffer.height != height:
            self._buffer.width = width
            self._buffer.height = height

    def _icon_size_changed_cb(self, image, pspec):
        self._buffer.icon_size = self.props.pixel_size

    def _icon_name_changed_cb(self, image, pspec):
        self._buffer.icon_name = self.props.icon_name

    def _file_changed_cb(self, image, pspec):
        self._buffer.file_name = self.props.file

    def do_get_preferred_height(self):
        '''Gtk widget implementation method'''
        self._sync_image_properties()
        surface = self._buffer.get_surface()
        if surface:
            height = surface.get_height()
        elif self._buffer.height:
            height = self._buffer.height
        else:
            height = 0
        return (height, height)

    def do_get_preferred_width(self):
        '''Gtk widget implementation method'''
        self._sync_image_properties()
        surface = self._buffer.get_surface()
        if surface:
            width = surface.get_width()
        elif self._buffer.width:
            width = self._buffer.width
        else:
            width = 0
        return (width, width)

    def do_draw(self, cr):
        '''Gtk widget implementation method'''
        self._sync_image_properties()
        sensitive = (self.is_sensitive())
        surface = self._buffer.get_surface(sensitive, self)
        if surface is None:
            return

        xpad, ypad = self.get_padding()
        xalign, yalign = self.get_alignment()
        requisition = self.get_child_requisition()
        if self.get_direction() != Gtk.TextDirection.LTR:
            xalign = 1.0 - xalign

        allocation = self.get_allocation()
        x = math.floor(xpad +
                       (allocation.width - requisition.width) * xalign)
        y = math.floor(ypad +
                       (allocation.height - requisition.height) * yalign)

        if self._scale != 1.0:
            cr.scale(self._scale, self._scale)

            margin = self._buffer.width * (1 - self._scale) / 2
            x, y = x + margin, y + margin

            x = x / self._scale
            y = y / self._scale

        cr.set_source_surface(surface, x, y)

        if self._alpha == 1.0:
            cr.paint()
        else:
            cr.paint_with_alpha(self._alpha)

    def set_xo_color(self, value):
        '''
        Set the colors used to display the icon

        Args:
            value (sugar3.graphics.xocolor.XoColor): new XoColor to use
        '''
        if self._buffer.xo_color != value:
            self._buffer.xo_color = value
            self.queue_draw()

    xo_color = GObject.Property(
        type=object, getter=None, setter=set_xo_color)
    '''
    icon.props.xo_color -> see :any:`set_xo_color`, note there is no getter
    '''

    def set_fill_color(self, value):
        '''
        Set the color used to fill the icon

        Args:
            value (str): SVG color string, like '#FFFFFF'
        '''
        if self._buffer.fill_color != value:
            self._buffer.fill_color = value
            self.queue_draw()

    def get_fill_color(self):
        '''
        Get the color used to fill the icon

        Returns:
            str, SVG color string, like '#FFFFFF'
        '''
        return self._buffer.fill_color

    fill_color = GObject.Property(
        type=object, getter=get_fill_color, setter=set_fill_color)
    '''
    icon.props.fill_color -> see :any:`get_fill_color`
        and :any:`set_fill_color`
    '''

    def set_stroke_color(self, value):
        '''
        Set the color used to paint the icon stroke

        Args:
            value (str): SVG color string, like '#282828'
        '''
        if self._buffer.stroke_color != value:
            self._buffer.stroke_color = value
            self.queue_draw()

    def get_stroke_color(self):
        '''
        Get the color used to paint the icon stroke

        Returns:
            str, SVG color string, like '#282828'
        '''
        return self._buffer.stroke_color

    stroke_color = GObject.Property(
        type=object, getter=get_stroke_color, setter=set_stroke_color)
    '''
    icon.props.stroke_color -> see :any:`get_stroke_color`
        and :any:`set_stroke_color`
    '''

    def set_badge_name(self, value):
        '''
        See the Badge Icons section at the top of the file.

        Args:
            value (str): the icon name for the badge
        '''
        if self._buffer.badge_name != value:
            self._buffer.badge_name = value
            self.queue_resize()

    def get_badge_name(self):
        '''
        Get the badge name, as set by :any:`set_badge_name`

        Returns:
            str, badge icon name
        '''
        return self._buffer.badge_name

    badge_name = GObject.Property(
        type=str, getter=get_badge_name, setter=set_badge_name)
    '''
    icon.props.badge_name -> see :any:`get_badge_name`
        and :any:`set_badge_name`
    '''

    def get_badge_size(self):
        '''
        Returns:
            int, size of badge icon, in pixels
        '''
        return int(_BADGE_SIZE * self.props.pixel_size)

    def set_alpha(self, value):
        '''
        Set the transparency for the icon.  Defaults to 1.0, which is
        fully visible icon.

        Args:
            value (float): alpha value from 0.0 to 1.0
        '''
        if self._alpha != value:
            self._alpha = value
            self.queue_draw()

    alpha = GObject.Property(
        type=float, setter=set_alpha)
    '''
    icon.props.alpha -> see :any:`set_alpha`, note no getter
    '''

    def set_scale(self, value):
        '''
        Scales the icon, with the transformation origin at the top left
        corner.  Note that this only scales the resulting drawing, so
        at large scales the icon will appear pixilated.

        Args:
            value (float): new scaling factor
        '''
        if self._scale != value:
            self._scale = value
            self.queue_draw()

    scale = GObject.Property(
        type=float, setter=set_scale)
    '''
    icon.props.scale -> see :any:`set_scale`, note no getter
    '''


class EventIcon(Gtk.EventBox):
    '''
    An Icon class that provides access to mouse events and that can act as a
    cursor-positioned palette invoker.

    The palette invoker can be used in 3 ways:

    1.  Set the palette during your constructor, see :any:`set_palette`
    2.  Override the create_palette method, see :any:`create_palette`
    3.  Set the tooltip, see :any:`create_tooltip`

    Otherwise, the icon setup api is the same as the basic :class:`Icon`.
    This EventIcon class supports the icon_name, stroke_color, fill_color,
    file_name, xo_color, pixel_size, scale and alpha keyword arguments as
    the :class:`Icon`.  The added arguments are as follows:

    Keyword Args:
        background_color (Gdk.Color): the color to draw the icon on top of.
            It defaults to None, which means no background is drawn
            (transparent).  The alpha channel of the Gdk.Color is disregarded.
        cache (bool): if True, the icon file contents will be cached to
            reduce disk usage
        palette (sugar3.graphics.palette.Palette): a palette to connect
    '''

    __gsignals__ = {
        'activate': (GObject.SignalFlags.RUN_FIRST, None, []),
    }

    __gtype_name__ = 'SugarEventIcon'

    def __init__(self, **kwargs):
        self._buffer = _IconBuffer()
        self._alpha = 1.0

        Gtk.EventBox.__init__(self)
        self.set_visible_window(False)
        self.set_above_child(True)
        self.add_events(Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.TOUCH_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK)
        # Connect after the default so that the palette can silence events
        # for example, after a touch palette invocation
        self.connect_after('button-release-event',
                           self.__button_release_event_cb)
        for key, value in six.iteritems(kwargs):
            self.set_property(key, value)

        from sugar3.graphics.palette import CursorInvoker
        self._palette_invoker = CursorInvoker()
        self._palette_invoker.attach(self)
        self.connect('destroy', self.__destroy_cb)

    def do_draw(self, cr):
        '''Gtk widget implementation method'''
        surface = self._buffer.get_surface()
        if surface:
            allocation = self.get_allocation()

            x = (allocation.width - surface.get_width()) / 2
            y = (allocation.height - surface.get_height()) / 2

            cr.set_source_surface(surface, x, y)
            if self._alpha == 1.0:
                cr.paint()
            else:
                cr.paint_with_alpha(self._alpha)

    def do_get_preferred_height(self):
        '''Gtk widget implementation method'''
        surface = self._buffer.get_surface()
        if surface:
            height = surface.get_height()
        elif self._buffer.height:
            height = self._buffer.height
        else:
            height = 0
        return (height, height)

    def do_get_preferred_width(self):
        '''Gtk widget implementation method'''
        surface = self._buffer.get_surface()
        if surface:
            width = surface.get_width()
        elif self._buffer.width:
            width = self._buffer.width
        else:
            width = 0
        return (width, width)

    def __destroy_cb(self, icon):
        if self._palette_invoker is not None:
            self._palette_invoker.detach()

    def set_file_name(self, value):
        if self._buffer.file_name != value:
            self._buffer.file_name = value
            self.queue_draw()

    def get_file_name(self):
        return self._buffer.file_name

    file_name = GObject.Property(
        type=object, getter=get_file_name, setter=set_file_name)

    def set_icon_name(self, value):
        if self._buffer.icon_name != value:
            self._buffer.icon_name = value
            self.queue_draw()

    def get_icon_name(self):
        return self._buffer.icon_name

    icon_name = GObject.Property(
        type=object, getter=get_icon_name, setter=set_icon_name)

    def set_xo_color(self, value):
        if self._buffer.xo_color != value:
            self._buffer.xo_color = value
            self.queue_draw()

    xo_color = GObject.Property(
        type=object, getter=None, setter=set_xo_color)

    def set_fill_color(self, value):
        if self._buffer.fill_color != value:
            self._buffer.fill_color = value
            self.queue_draw()

    def get_fill_color(self):
        return self._buffer.fill_color

    fill_color = GObject.Property(
        type=object, getter=get_fill_color, setter=set_fill_color)

    def set_stroke_color(self, value):
        if self._buffer.stroke_color != value:
            self._buffer.stroke_color = value
            self.queue_draw()

    def get_stroke_color(self):
        return self._buffer.stroke_color

    stroke_color = GObject.Property(
        type=object, getter=get_stroke_color, setter=set_stroke_color)

    def set_background_color(self, value):
        '''
        Args:
            value (Gdk.Color): color use as background (alpha is ignored),
                or None meaning no background is drawn (transparent)
        '''
        if self._buffer.background_color != value:
            self._buffer.background_color = value
            self.queue_draw()

    def get_background_color(self):
        '''
        Returns:
            Gdk.Color, current background color, may be None
        '''
        return self._buffer.background_color

    background_color = GObject.Property(
        type=object, getter=get_background_color, setter=set_background_color)
    '''
    event_icon.props.get_background_color -> see :any:`set_background_color`
        and :any:`get_background_color`
    '''

    def set_size(self, value):
        if self._buffer.width != value:
            self._buffer.width = value
            self._buffer.height = value
            self.queue_resize()

    def get_size(self):
        return self._buffer.width

    pixel_size = GObject.Property(
        type=object, getter=get_size, setter=set_size)

    def set_scale(self, value):
        if self._buffer.scale != value:
            self._buffer.scale = value
            self.queue_resize()

    def get_scale(self):
        return self._buffer.scale

    scale = GObject.Property(
        type=float, getter=get_scale, setter=set_scale)

    def set_alpha(self, alpha):
        if self._alpha != alpha:
            self._alpha = alpha
            self.queue_draw()

    alpha = GObject.Property(
        type=float, setter=set_alpha)

    def set_cache(self, value):
        '''
        Sugar caches icon file contents in a smart cache.  Currently, we use
        a LRU (Least Recently Used) algorithm to manage the cache.

        Args:
            value (bool): if True, the icon file will be cached in the LRU
        '''
        self._buffer.cache = value

    def get_cache(self):
        '''
        Returns:
            bool, if the icon file will be saved in the LRU
        '''
        return self._buffer.cache

    cache = GObject.Property(
        type=bool, default=False, getter=get_cache, setter=set_cache)
    '''
    event_icon.props.cache -> see :any:`set_cache` and :any:`get_cache`
    '''

    def set_badge_name(self, value):
        if self._buffer.badge_name != value:
            self._buffer.badge_name = value
            self.queue_draw()

    def get_badge_name(self):
        return self._buffer.badge_name

    badge_name = GObject.Property(
        type=object, getter=get_badge_name, setter=set_badge_name)

    def create_palette(self):
        '''
        The create_palette function is called when the palette needs to be
        invoked.  For example, when the user has right clicked the icon or
        the user has hovered over the icon for a long time.

        The create_palette will only be called once or zero times.  The palette
        returned will be stored and re-used if the user invokes the palette
        multiple times.

        Your create_palette implementation does not need to
        :any:`Gtk.Widget.show` the palette, as this will be done by the
        invoker.  However, you still need to show
        the menu items, etc that you place in the palette.

        Returns:
            sugar3.graphics.palette.Palette, or None to indicate that you
            do not want a palette shown

        The default implementation returns None, to indicate no palette should
        be shown.
        '''
        return None

    def get_palette(self):
        '''
        Gets the current palette, either set by :any:`set_palette` or cached
        after a call to :any:`create_palette`
        '''
        return self._palette_invoker.palette

    def set_palette(self, palette):
        '''
        Sets the palette to show.  If the palette is not None, this will
        override the palette set by create_palette.

        Args:
            palette (sugar3.graphics.palette.Palette): palette or None
        '''
        self._palette_invoker.palette = palette

    palette = GObject.Property(
        type=object, setter=set_palette, getter=get_palette)
    '''
    event_icon.props.palette -> see :any:`get_palette` and :any:`set_palette`
    '''

    def get_palette_invoker(self):
        return self._palette_invoker

    def set_palette_invoker(self, palette_invoker):
        self._palette_invoker.detach()
        self._palette_invoker = palette_invoker

    palette_invoker = GObject.Property(
        type=object, setter=set_palette_invoker, getter=get_palette_invoker)

    def set_tooltip(self, text):
        '''
        Creates a palette with the tooltip text.  This will override any
        current palette set through :any:`set_palette` or that will ever be
        returned by :any:`create_palette`.

        Args:
            text (str): tooltip text
        '''
        from sugar3.graphics.palette import Palette

        self.set_palette(Palette(text))

    def __button_release_event_cb(self, icon, event):
        if event.button == 1:
            alloc = self.get_allocation()
            if 0 < event.x < alloc.width and 0 < event.y < alloc.height:
                self.emit('activate')


class CanvasIcon(EventIcon):
    '''
    An EventIcon with active and prelight states, and a styleable
    background.  If the icon pops up a palette, the prelight state is
    set until the palette pops down.  This is used to render a light
    grey highlight, however can be configured by Gtk+ CSS with the
    `:prelight` selector.

    Care should to use :any:`connect_to_palette_pop_events` for all palettes
    created and shown around this icon.
    '''

    __gtype_name__ = 'SugarCanvasIcon'

    def __init__(self, **kwargs):
        EventIcon.__init__(self, **kwargs)
        self._button_down = False

        self.connect('enter-notify-event', self.__enter_notify_event_cb)
        self.connect('leave-notify-event', self.__leave_notify_event_cb)
        self.connect('button-press-event', self.__button_press_event_cb)
        self.connect('button-release-event', self.__button_release_event_cb)

    def connect_to_palette_pop_events(self, palette):
        '''
        Connect to the palette's popup and popdown events, so that the prelight
        state is set at the right times.  You should run this call
        before you :any:`EventIcon.set_palette` or before you return from
        your :any:`EventIcon.create_palette` function, eg:

            def create_palette(self):
                palette = ...

                self.connect_to_palette_pop_events(palette)
                return palette

        Args:
            palette (sugar3.graphics.palette.Palette): palette to connect
        '''
        palette.connect('popup', self.__palette_popup_cb)
        palette.connect('popdown', self.__palette_popdown_cb)

    def do_draw(self, cr):
        '''Gtk widget implementation method'''
        allocation = self.get_allocation()
        context = self.get_style_context()
        Gtk.render_background(context, cr, 0, 0,
                              allocation.width,
                              allocation.height)

        EventIcon.do_draw(self, cr)

    def __enter_notify_event_cb(self, icon, event):
        self.set_state_flags(self.get_state_flags() | Gtk.StateFlags.PRELIGHT,
                             clear=True)
        if self._button_down:
            self.set_state_flags(Gtk.StateFlags.ACTIVE, clear=False)

    def __leave_notify_event_cb(self, icon, event):
        if self.palette and self.palette.is_up():
            return

        self.unset_state_flags(Gtk.StateFlags.PRELIGHT | Gtk.StateFlags.ACTIVE)

    def __button_press_event_cb(self, icon, event):
        if self.palette and not self.palette.is_up():
            self._button_down = True
            self.set_state_flags(
                self.get_state_flags() | Gtk.StateFlags.ACTIVE,
                clear=True)

    def __button_release_event_cb(self, icon, event):
        self.unset_state_flags(Gtk.StateFlags.ACTIVE)
        self._button_down = False

    def __palette_popup_cb(self, palette):
        self.set_state_flags(Gtk.StateFlags.PRELIGHT, clear=True)

    def __palette_popdown_cb(self, palette):
        self.unset_state_flags(Gtk.StateFlags.PRELIGHT)


if hasattr(CanvasIcon, 'set_css_name'):
    CanvasIcon.set_css_name('canvasicon')


class CellRendererIcon(Gtk.CellRenderer):

    __gtype_name__ = 'SugarCellRendererIcon'

    __gsignals__ = {
        'clicked': (GObject.SignalFlags.RUN_FIRST, None, [object]),
    }

    def __init__(self, treeview=None):
        # treeview is not used anymore, is here just to not break the API
        if treeview is not None:
            logging.warning('CellRendererIcon: treeview parameter in '
                            'constructor is deprecated')
        self._buffer = _IconBuffer()
        self._buffer.cache = True
        self._xo_color = None
        self._fill_color = None
        self._stroke_color = None
        self._prelit_fill_color = None
        self._prelit_stroke_color = None
        self._active_state = False
        self._cached_offsets = None

        Gtk.CellRenderer.__init__(self)

        self._is_scrolling = False

    def connect_to_scroller(self, scrolled):
        scrolled.connect('scroll-start', self._scroll_start_cb)
        scrolled.connect('scroll-end', self._scroll_end_cb)

    def _scroll_start_cb(self, event):
        self._is_scrolling = True

    def _scroll_end_cb(self, event):
        self._is_scrolling = False

    def is_scrolling(self):
        return self._is_scrolling

    def create_palette(self):
        return None

    def set_file_name(self, value):
        if self._buffer.file_name != value:
            self._buffer.file_name = value

    file_name = GObject.Property(type=str, setter=set_file_name)

    def set_icon_name(self, value):
        if self._buffer.icon_name != value:
            self._buffer.icon_name = value

    icon_name = GObject.Property(type=str, setter=set_icon_name)

    def get_xo_color(self):
        return self._xo_color

    def set_xo_color(self, value):
        self._xo_color = value

    xo_color = GObject.Property(type=object,
                                getter=get_xo_color, setter=set_xo_color)

    def set_fill_color(self, value):
        if self._fill_color != value:
            self._fill_color = value

    fill_color = GObject.Property(type=object, setter=set_fill_color)

    def set_stroke_color(self, value):
        if self._stroke_color != value:
            self._stroke_color = value

    stroke_color = GObject.Property(type=object, setter=set_stroke_color)

    def set_prelit_fill_color(self, value):
        if self._prelit_fill_color != value:
            self._prelit_fill_color = value

    prelit_fill_color = GObject.Property(type=object,
                                         setter=set_prelit_fill_color)

    def set_prelit_stroke_color(self, value):
        if self._prelit_stroke_color != value:
            self._prelit_stroke_color = value

    prelit_stroke_color = GObject.Property(type=object,
                                           setter=set_prelit_stroke_color)

    def set_background_color(self, value):
        if self._buffer.background_color != value:
            self._buffer.background_color = value

    background_color = GObject.Property(type=object,
                                        setter=set_background_color)

    def set_size(self, value):
        if self._buffer.width != value:
            self._buffer.width = value
            self._buffer.height = value

            self._cached_offsets = None

    size = GObject.Property(type=object, setter=set_size)

    def do_get_size(self, widget, cell_area, x_offset=None, y_offset=None,
                    width=None, height=None):
        width = self._buffer.width + self.props.xpad * 2
        height = self._buffer.height + self.props.ypad * 2
        xoffset = 0
        yoffset = 0

        if width > 0 and height > 0 and cell_area is not None:

            if widget.get_direction() == Gtk.TextDirection.RTL:
                xoffset = 1.0 - self.props.xalign
            else:
                xoffset = self.props.xalign

            xoffset = max(xoffset * (cell_area.width - width), 0)
            yoffset = max(self.props.yalign * (cell_area.height - height), 0)
            self._cached_offsets = xoffset, yoffset

        return xoffset, yoffset, width, height

    def _get_offsets(self, widget, cell_area):
        if self._cached_offsets is not None:
            return self._cached_offsets

        xoffset, yoffset, width_, height_ = self.do_get_size(widget, cell_area)
        return xoffset, yoffset

    def do_activate(self, event, widget, path, background_area, cell_area,
                    flags):
        pass

    def do_start_editing(self, event, widget, path, background_area, cell_area,
                         flags):
        pass

    def do_render(self, cr, widget, background_area, cell_area, flags):
        if not self._is_scrolling:

            context = widget.get_style_context()
            context.save()
            context.add_class("sugar-icon-cell")

            def is_pointer_inside():
                # widget is the treeview
                x, y = widget.get_pointer()
                x, y = widget.convert_widget_to_bin_window_coords(x, y)
                return ((cell_area.x <= x <= cell_area.x + cell_area.width) and
                        (cell_area.y <= y <= cell_area.y + cell_area.height))

            pointer_inside = is_pointer_inside()

            # The context will have prelight state if the mouse pointer is
            # in the entire row, but we want that state if the pointer is
            # in this cell only:
            if flags & Gtk.CellRendererState.PRELIT:
                if pointer_inside:
                    if self._active_state:
                        context.set_state(Gtk.StateFlags.ACTIVE)
                else:
                    context.set_state(Gtk.StateFlags.NORMAL)

            Gtk.render_background(
                context, cr, background_area.x, background_area.y,
                background_area.width, background_area.height)

            if self._xo_color is not None:
                stroke_color = self._xo_color.get_stroke_color()
                fill_color = self._xo_color.get_fill_color()
                prelit_fill_color = None
                prelit_stroke_color = None
            else:
                stroke_color = self._stroke_color
                fill_color = self._fill_color
                prelit_fill_color = self._prelit_fill_color
                prelit_stroke_color = self._prelit_stroke_color

            has_prelit_colors = None not in [prelit_fill_color,
                                             prelit_stroke_color]

            if flags & Gtk.CellRendererState.PRELIT and has_prelit_colors and \
                    pointer_inside:

                self._buffer.fill_color = prelit_fill_color
                self._buffer.stroke_color = prelit_stroke_color
            else:
                self._buffer.fill_color = fill_color
                self._buffer.stroke_color = stroke_color
        else:
            if self._xo_color is not None:
                self._buffer.fill_color = self._xo_color.get_fill_color()
                self._buffer.stroke_color = self._xo_color.get_stroke_color()
            else:
                self._buffer.fill_color = self._fill_color
                self._buffer.stroke_color = self._stroke_color

        surface = self._buffer.get_surface()
        if surface is None:
            return

        xoffset, yoffset = self._get_offsets(widget, cell_area)

        x = cell_area.x + xoffset
        y = cell_area.y + yoffset

        cr.set_source_surface(surface, math.floor(x), math.floor(y))
        cr.rectangle(cell_area.x, cell_area.y, cell_area.width,
                     cell_area.height)
        cr.clip()
        cr.paint()


def get_icon_state(base_name, perc, step=5):
    '''
    Get the closest icon name for a given state in percent.

    First, you need a set of icons. They must be prefixed with `base_name`,
    for example "network-wireless". They must be suffixed with 3 digit
    percentage numbers, for example "-000", "-200", etc. Eventually, you get
    a collection of icon names like:

    * network-wireless-000
    * network-wireless-020
    * network-wireless-040
    * network-wireless-060
    * network-wireless-080
    * network-wireless-100

    All of these icons must be placed in the icon theme, such that they are
    addressable by their `icon_name`.

    Args:
        base_name (str): base icon name, eg `network-wireless`
        perc (float): desired percentage between 0 and 100, eg. 67.8

    Keyword Arguments:
        step (int): step to increment to find all possible icons

            From the example above, we could step 5, because 0, 5, 10, 15, etc,
            includes all number suffixes in our set of icons (0, 20, 40, etc).

            If we had the number suffixes 0, 33, 66, 99, we could not use 5,
            as none of the numbers are divisible by 5.

    Returns:
        str, icon name that represent given state, or None if not found
    '''
    strength = round(perc / step) * step
    icon_theme = Gtk.IconTheme.get_default()

    while strength <= 100 and strength >= 0:
        icon_name = '%s-%03d' % (base_name, strength)
        if icon_theme.has_icon(icon_name):
            return icon_name

        strength = strength + step


def get_icon_file_name(icon_name):
    '''
    Resolves a given icon name into a file path.  Looks for any icon in them
    theme, including those in sugar-artwork and those in the activities
    '/icons/' directory.

    Returns:
        str, path to icon, or None is the icon is not found in the theme
    '''
    icon_theme = Gtk.IconTheme.get_default()
    info = icon_theme.lookup_icon(icon_name, Gtk.IconSize.LARGE_TOOLBAR, 0)
    if not info:
        return None
    filename = info.get_filename()
    del info
    return filename


def get_surface(**kwargs):
    '''
    Get cairo surface of the icon.  Supports the same arguments as
    :any:`Icon`, in exactly the same way.

    Returns:
        cairo surface or None if image was not found
    '''
    icon = _IconBuffer()
    for key, value in list(kwargs.items()):
        icon.__setattr__(key, value)
    return icon.get_surface()



# Copyright (C) 2013 Daniel Narvaez
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

import json
import os
import logging

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
gi.require_version('WebKit2', '4.0')

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import WebKit2
from gi.repository import Gtk
from gi.repository import GdkX11
assert GdkX11

from sugar3.graphics.objectchooser import ObjectChooser
from gi.repository import SugarExt
from sugar3.activity import activity


class FilePicker(ObjectChooser):
    def __init__(self, parent):
        ObjectChooser.__init__(self, parent)

    def run(self):
        jobject = None
        _file = None
        try:
            result = ObjectChooser.run(self)
            if result == Gtk.ResponseType.ACCEPT:
                jobject = self.get_selected_object()
                logging.debug('FilePicker.run: %r', jobject)

                if jobject and jobject.file_path:
                    _file = jobject.file_path
                    logging.debug('FilePicker.run: file=%r', _file)
        finally:
            if jobject is not None:
                jobject.destroy()

        return _file


class WebActivity(Gtk.Window):
    def __init__(self, handle):
        Gtk.Window.__init__(self)

        self._activity_id = handle.activity_id
        self._object_id = handle.object_id
        self._bundle_id = os.environ["SUGAR_BUNDLE_ID"]
        self._bundle_path = os.environ["SUGAR_BUNDLE_PATH"]
        self._inspector_visible = False

        self.set_decorated(False)
        self.maximize()

        self.connect("key-press-event", self._key_press_event_cb)
        self.connect('realize', self._realize_cb)
        self.connect('destroy', self._destroy_cb)

        context = WebKit2.WebContext.get_default()
        context.register_uri_scheme("activity", self._app_scheme_cb, None)

        self._web_view = WebKit2.WebView()
        self._web_view.connect("load-changed", self._loading_changed_cb)
        self._web_view.connect('run-file-chooser', self.__run_file_chooser)

        self.add(self._web_view)
        self._web_view.show()

        settings = self._web_view.get_settings()
        settings.set_property("enable-developer-extras", True)

        self._web_view.load_uri("activity://%s/index.html" % self._bundle_id)

        self.set_title(activity.get_bundle_name())

    def run_main_loop(self):
        Gtk.main()

    def _realize_cb(self, window):
        xid = window.get_window().get_xid()
        SugarExt.wm_set_bundle_id(xid, self._bundle_id)
        SugarExt.wm_set_activity_id(xid, str(self._activity_id))

    def _destroy_cb(self, window):
        self.destroy()
        Gtk.main_quit()

    def _loading_changed_cb(self, web_view, load_event):
        if load_event == WebKit2.LoadEvent.FINISHED:
            key = os.environ["SUGAR_APISOCKET_KEY"]
            port = os.environ["SUGAR_APISOCKET_PORT"]

            env_json = json.dumps({"apiSocketKey": key,
                                   "apiSocketPort": port,
                                   "activityId": self._activity_id,
                                   "bundleId": self._bundle_id,
                                   "objectId": self._object_id,
                                   "activityName": activity.get_bundle_name()})

            script = """
                     var environment = %s;

                     if (window.sugar === undefined) {
                         window.sugar = {};
                     }

                     window.sugar.environment = environment;

                     if (window.sugar.onEnvironmentSet)
                         window.sugar.onEnvironmentSet();
                    """ % env_json

            self._web_view.run_javascript(script, None, None, None)

    def __run_file_chooser(self, browser, request):
        picker = FilePicker(self)
        chosen = picker.run()
        picker.destroy()

        if chosen:
            request.select_files([chosen])
        else:
            request.cancel()
        return True

    def _key_press_event_cb(self, window, event):
        key_name = Gdk.keyval_name(event.keyval)

        if event.get_state() & Gdk.ModifierType.CONTROL_MASK and \
           event.get_state() & Gdk.ModifierType.SHIFT_MASK:
            if key_name == "I":
                inspector = self._web_view.get_inspector()
                if self._inspector_visible:
                    inspector.close()
                    self._inspector_visible = False
                else:
                    inspector.show()
                    self._inspector_visible = True

                return True

    def _app_scheme_cb(self, request, user_data):
        path = os.path.join(self._bundle_path,
                            os.path.relpath(request.get_path(), "/"))

        request.finish(Gio.File.new_for_path(path).read(None),
                       -1, Gio.content_type_guess(path, None)[0])
